# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt
from hysteresis.plotSpecial import CycleViewer
# from pytest import MonkeyPatch

disp = np.loadtxt('UFP_Disp.out',delimiter=' ')
force = np.loadtxt('UFP_RFrc.out',delimiter=' ')
testHys2 = np.column_stack([disp[:,1], -force[:,1]])

DamperHys = hys.Hysteresis(testHys2) 
# DamperHys.plot(True)


def test_CycleViewer(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)

    CycleViewer(DamperHys, xlims = [0,20])
    
    assert True == True


if __name__ == "__main__":
    
    test_CycleViewer()


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt
import scipy
from scipy.interpolate import interp1d

# a triangle with small reversals
x = np.linspace(0, 1, 1000)*10
triangleBig = scipy.signal.sawtooth(x*2,0.5)
triangleSmall = scipy.signal.sawtooth(x*20,0.5)/7
triangle = triangleBig + triangleSmall
trianglexy = np.column_stack((x,triangle))
notchedTriangles = hys.SimpleCurve(trianglexy, findPeaks = True)

def test_notchedTriangles_peaks():
    notchedTriangles.setPeaks()
    peaks = notchedTriangles.peakIndexes

    test1 = peaks[4] == 62
    test2 = peaks[14] == 220
    test3 = peaks[31] == 486
    
    assert np.all([test1,test2,test3])

def test_notchedTriangles_Slope():
    notchedTriangles.setSlope()
    slope = notchedTriangles.slope
   
    assert abs(slope[-1] - -0.54567409060066) < 10**-8


def test_notchedTriangles_Area():

    notchedTriangles.setArea()
    Area = notchedTriangles.getNetArea()

    assert abs(Area - -0.3595804712700046) < 10**-8

def test_notchedTriangles_recalc_peaks():
    
    notchedTriangles = hys.SimpleCurve(trianglexy, findPeaks = True)
    notchedTriangles.recalculatePeaks(peakProminence = 0.8)
    peaks2 = notchedTriangles.peakIndexes

    assert peaks2[2] == 314


# -*- coding: utf-8 -*-
"""
Created on Fri May  7 20:15:19 2021

@author: Christian
"""

import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.widgets import Button

import numpy as np

# Add this function to somewhere else
def init_Animation():
    fig, ax = plt.subplots()
    
    return fig, ax
    


def getAnixy(rawFrames, skipFrames):
    xyAni = rawFrames[::skipFrames,:]
    
    return xyAni

def getAniFrames(x, targetdx):
    """
    Returns a frame every target dx. Can be useful for pre-processing data
    if input data has a variable timestep.
    
    No linearl inerpolatin is used for intermediate frames.
    """
    NFrames = len(x)   
    NframesOut = []
    
    jj = 0
    for ii in range(NFrames):
        
        while jj*targetdx < x[ii]:
            NframesOut.append(x[ii])
            jj+=1

    return np.array(NframesOut)

class AnimationBase:    
    
    def __init__(self):
        self.play = True
        
        # Replace with imported function
        fig, ax = init_Animation()
    
        # Connect clicking
        # fig.canvas.mpl_connect('button_press_event', self.on_click)
        fig.canvas.mpl_connect('button_press_event', self.toggle_pause)
        
        self.fig = fig
        self.ax = ax            
    
    
    
    
    def togglePlay(self):
        self.play = self.play == False
        
        
    def on_click(self, event):
        xclick = event.x
        yclick = event.y
        
        return xclick, yclick
        # print(xclick, yclick)
    
    # To be overwritten
    def toggle_pause(self,event, *args, **kwargs):
        pass
        
   #      # Check where the click happened
   #      (xm,ym),(xM,yM) = plotSlider.label.clipbox.get_points()
   #      if xm < event.x < xM and ym < event.y < yM:
   #          # Event happened within the slider, ignore since it is handled in update_slider
   #          return
   #      else:
   #          # Toggle on off based on clicking
   #          global is_paused
   #          if is_paused == True:
   #              is_paused=False
   #          elif is_paused == False:
   #              is_paused=True



# class FrameHelper():
    
#     def __init__(self, pointsPerFrame = 1, skipFrames = 1, skipStart = 0, skipEnd = 0):
    
#         self.pointsPerFrame = pointsPerFrame        




class Animation(AnimationBase):
    
    def __init__(self, Curve, pointsPerFrame = 1, skipFrames = 1, skipStart = 0, skipEnd = 0, interval = 50):
        
        super().__init__()
        
        self.Curve = Curve
        self.xy = Curve.xy
        
        self.pointsPerFrame = pointsPerFrame        
        self.interval = interval
        
        xyAni = getAnixy(Curve.xy, skipFrames)
        
        self.xyAni = xyAni
        # self.xyAni = self.xy 
        self.Nframes = int(len(self.xyAni) / pointsPerFrame)
        self.frames = np.arange(self.Nframes)
        
        self.lines = []
        
        
        # self.fig.canvas.mpl_connect('button_press_event', self.toggle_pause)


    # def setAniXY()
    def initfunc(self):
        line = plt.plot(self.xyAni[:,0], self.xyAni[:,1])[0]
        self.lines.append(line)    # def initAnimation(self):
        
        return line,
    
    def update(self, frame):
        
        # for ii in range()
        points = int(frame*self.pointsPerFrame)
        
        newXY = self.xyAni[:points,:]
        line = self.lines[0]
        line.set_data(newXY[:,0], newXY[:,1])
        
        return line,
    
    def Animate(self):
        self.ani = animation.FuncAnimation(self.fig, self.update, self.frames, self.initfunc, 
                            interval=self.interval, blit=True)


    # def toggle_pause(self, *args, **kwargs):
    #     self.togglePlay()
        
    #     if self.play:
    #         self.ani.resume()
    #     else:
    #         self.ani.pause()   
    
class JointAnimation(AnimationBase):
    
    def __init__(self, Curves, pointsPerFrame = 1, skipFrames = 1, skipStart = 0, skipEnd = 0, interval = 50):
        
        super().__init__()
        
        self.pointsPerFrame = pointsPerFrame        
        self.interval = interval
        
        self.Curves = Curves
        self.Ncurves = len(Curves)
        
        xyAni = getAnixy(Curves[0].xy, skipFrames)
        self.Nframes = int(len(xyAni) / pointsPerFrame)
        self.frames = np.arange(self.Nframes)        
        
        
        self.xyCurves = [None]*self.Ncurves
        self.lines = []
        
        for ii in range(self.Ncurves):
            curve = self.Curves[ii]
            xy = curve.xy
        
            xyAni = getAnixy(xy, skipFrames)
            self.xyCurves[ii] = xyAni
        
            # self.xyAni = xyAni

        
        # self.fig.canvas.mpl_connect('button_press_event', self.toggle_pause)


    # def setAniXY()
    def initfunc(self):
        
        for ii in range(self.Ncurves):
            tempXY = self.xyCurves[ii]
            line = plt.plot(tempXY[:,0], tempXY[:,1])[0]
            self.lines.append(line)    # def initAnimation(self):
        
        return self.lines
    
    def update(self, frame):
        
        # print(frame)
        points = int(frame*self.pointsPerFrame)
        # print(points)
        lines = [None]*self.Ncurves
        for ii in range(self.Ncurves):

            tempXY = self.xyCurves[ii]
            # print()
            
            newXY = tempXY[:points,:]
            line = self.lines[ii]
            line.set_data(newXY[:,0], newXY[:,1])
            lines[ii] = line
            
            
            # lines[ii] = line
        # lines = self.lines
        return lines
    
    def Animate(self):
        self.ani = animation.FuncAnimation(self.fig, self.update, self.frames, self.initfunc, 
                            interval=50, blit=True)   
    
    
    
    
    
    
    
 
        
    # axSlider = plt.axes([0.25, .03, 0.50, 0.02])
    # plotSlider = Slider(axSlider, 'Time', framesTime[FrameStart], framesTime[FrameEnd], valinit=framesTime[FrameStart])        
        
        
        
   # # Slider Location and size relative to plot
   #  # [x, y, xsize, ysize]
   #  axSlider = plt.axes([0.25, .03, 0.50, 0.02])
   #  plotSlider = Slider(axSlider, 'Time', framesTime[FrameStart], framesTime[FrameEnd], valinit=framesTime[FrameStart])
    
   #  # Animation controls
   #  global is_paused
   #  is_paused = False # True if user has taken control of the animation   
    
   #  def on_click(event):
   #      # Check where the click happened
   #      (xm,ym),(xM,yM) = plotSlider.label.clipbox.get_points()
   #      if xm < event.x < xM and ym < event.y < yM:
   #          # Event happened within the slider, ignore since it is handled in update_slider
   #          return
   #      else:
   #          # Toggle on off based on clicking
   #          global is_paused
   #          if is_paused == True:
   #              is_paused=False
   #          elif is_paused == False:
   #              is_paused=True
                
   #  def animate2D_slider(Time):
   #      """
   #      The slider value is liked with the plot - we update the plot by updating
   #      the slider.
   #      """
   #      global is_paused
   #      is_paused=True
   #      # Convert time to frame
   #      TimeStep = (np.abs(framesTime - Time)).argmin()
               
   #      # The current node coordinants in (x,y) or (x,y,z)
   #      CurrentNodeCoords =  nodes[:,1:] + Disp[TimeStep,:,:]
   #      # Update Plots
        
   #      # update node locations
   #      EqfigNodes.set_xdata(CurrentNodeCoords[:,0]) 
   #      EqfigNodes.set_ydata(CurrentNodeCoords[:,1])
           
   #      # Get new node mapping
   #      # I don't like doing this loop every time - there has to be a faster way
   #      xy_labels = {}
   #      for jj in range(Nnodes):
   #          xy_labels[nodeLabels[jj]] = CurrentNodeCoords[jj,:]
        
   #      # Define the surface
   #      SurfCounter = 0
        
   #      # update element locations
   #      for jj in range(Nele):
   #          # Get the node number for the first and second node connected by the element
   #          TempNodes = elements[jj][1:]
   #          # This is the xy coordinates of each node in the group
   #          TempNodeCoords = [xy_labels[node] for node in TempNodes] 
   #          coords_x = [xy[0] for xy in TempNodeCoords]
   #          coords_y = [xy[1] for xy in TempNodeCoords]
            
   #          # Update element lines    
   #          EqfigLines[jj].set_xdata(coords_x)
   #          EqfigLines[jj].set_ydata(coords_y)
   #          # print('loop start')
   #          # Update the surface if necessary
   #          if 2 < len(TempNodes):
   #              tempxy = np.column_stack([coords_x, coords_y])
   #              EqfigSurfaces[SurfCounter].xy = tempxy
   #              SurfCounter += 1
       
   #      # update time Text
   #      # time_text.set_text("Time= "+'%.2f' % time[TimeStep]+ " s")
        
   #      # redraw canvas while idle
   #      fig.canvas.draw_idle()    
            
   #      return EqfigNodes, EqfigLines, EqfigSurfaces, EqfigText        
        
        
        
	# Saving
    # if Movie != "none":
    #     MovefileName = Movie + '.mp4'
    #     ODBdir = Model+"_ODB"		# ODB Dir name
    #     Movfile = os.path.join(ODBdir, LoadCase, MovefileName)
    #     print("Saving the animation movie as "+MovefileName+" in "+ODBdir+"->"+LoadCase+" folder")
    #     ani.save(Movfile, writer='ffmpeg')             
        
def init_gui():
    fig, ax = plt.subplots()
    
    return fig, ax        



class GUIBase:
    

    def __init__(self):
        fig, ax = init_gui()
        self.fig = fig
        self.ax = ax
        
        
    def addNextBtns(self):
        
        self.axprev = plt.axes([0.7, 0.05, 0.1, 0.075])
        self.axnext = plt.axes([0.81, 0.05, 0.1, 0.075])
        self.bnext = Button(self.axnext, 'Next')
        self.bnext.on_clicked(self.plotNext)
        self.bprev = Button(self.axprev, 'Previous')
        self.bprev.on_clicked(self.plotPrev)
        






class CycleViewer(GUIBase):

    
    
    def __init__(self, hysteresis, showCumulative = False):
        
        super().__init__()
        
        self.showCumulative = showCumulative 
        
        
        self.hysteresis = hysteresis
        self.curves = np.array(hysteresis.Cycles)
        
        self.ind = 0
        self.Ncurves = hysteresis.NCycles
        # xyAni = getAnixy(Curve.xy, skipFrames)
        
        # add line
        self.line = self.curves[-1].plot()
        
        # a list of active curves
        # self.activeLines =  [self.line]
        
        
        # adjsut plot and add buttons
        plt.subplots_adjust(bottom=0.2)
        self.addNextBtns()
        
        plt.show()
    
    # def getActiveCurves    
    
    
    
    # def initAnimation(self):
    def updateplt(self, ind):
        # print('pressed')
        
        # xy = []
        # for ii in range(ind):
        #     xy.append(self.curves[ind].xy)
        
        # xy = np.concatenate(xy)
        xy = self.curves[ind].xy
        self.line.set_xdata(xy[:,0])
        self.line.set_ydata(xy[:,1])
        plt.draw()
        
    # def updatepltCumulative
        
    def plotNext(self, event):
        self.ind = (self.ind + 1) % self.Ncurves
        self.updateplt(self.ind)

        
    def plotPrev(self, event):
        self.ind = (self.ind - 1) % self.Ncurves
        self.updateplt(self.ind)


class CycleViewerCum(GUIBase):

    plotMethods = {}
    
    def __init__(self, hysteresis, plotMethod = 0):
        """
        An interactive  figures to be plotted 
        
        
        Plots must use a qt backend for the GUI elements to work
        correctly. %matplotlib qt
        

        Parameters
        ----------
        hysteresis : TYPE
            DESCRIPTION.
        plotMethod : TYPE, optional
            DESCRIPTION. The default is 'single'.

        Returns
        -------
        None.

        """
        
        
        
        super().__init__()
        
        # Chhose the plot method
        self.plotMethod = plotMethod 
        plotMethods = [self.getSingleXy, self.getCumXy]
        self.getPlotXY = plotMethods[plotMethod]
        
        # Store the data.
        self.hysteresis = hysteresis
        self.curves = np.array(hysteresis.Cycles)
        # print(self.curves)
        
        # self.ind = np.array([0],dtype=int)
        self.ind = 0
        self.Ncurves = hysteresis.NCycles
        # xyAni = getAnixy(Curve.xy, skipFrames)
        
        # add initial lines and text
        self.line = self.curves[-1].plot()
        self.annot = self.ax.annotate(self.ind,(0.05,0.05), xycoords='axes fraction')        
        
        # adjsut plot and add buttons
        plt.subplots_adjust(bottom=0.2)
        self.addNextBtns()
        
        # start plot
        plt.show()
        # self.updateplt(0)   
    
    
    def getSingleXy(self, ind):
        curve = self.curves[ind]
            
        return curve.xy

    
    def getCumXy(self,ind):
        curves = self.curves[:ind+1]
        xy = np.array([])
        xy.shape = (0,2)

        for curve in curves:
            
            tempxy = curve.xy
            xy = np.concatenate([xy,tempxy])
            
        return xy
    
    def updateplt(self, ind):

        xy = self.getCumXy(ind)

        self.line.set_xdata(xy[:,0])
        self.line.set_ydata(xy[:,1])
        
        text = f'Cycle {self.ind}'

        self.annot.set_text(text)
        
        
        plt.draw()
        
        
    def plotNext(self, event):
        self.ind = (self.ind + 1) % self.Ncurves
        self.updateplt(self.ind)

        
    def plotPrev(self, event):
        self.ind = (self.ind - 1) % self.Ncurves
        self.updateplt(self.ind)



    # def CyclePlot()
    
            # def initAnimation(self):
# def updateplt(self, ind):
#     print('pressed')
#     xy = self.curves[ind].xy
#     self.line.set_xdata(xy[:,0])
#     self.line.set_ydata(xy[:,1])
#     plt.draw()
    
# def plotNext(self, event):
#     print('pressed')
#     self.ind = (self.ind + 1) % self.Ncurves
#     self.updateplt(self.ind)

    
# def plotPrev(self, event):
#     self.ind = (self.ind - 1) % self.Ncurves
#     self.updateplt(self.ind)
        
        



from .curve import Curve, Hysteresis
import hysteresis.env as env
from .resample import resample


# =============================================================================
# Compare
# =============================================================================
# TODO: consider storing the sample function within the class itself.

def compareCycle(Curve1:Curve, Curve2:Curve, Nsample = 10):
    """
    Compares two Curve objects by resampling them using linear interplation,
    then comparing the distance between both cycles in a comon domain.

    Parameters
    ----------
    Curve1 : Curve
        The first curve, must be non-Hysteresis.
    Curve2 : Curve
        The second curve, must be non-Hysteresis.
    Nsample : int
        The number of samples that the comparison curves will have.
    sampleFunction : function, optional
        The sample function to be used to compare the curves. 
        The default function is "defaultSampleFunction".

    Returns
    -------
    diff : float
        The net difference between each curve.

    """
    sampleFunction = env.environment.fSample
    
    
    if Curve1.Npoints != Curve1.Npoints:
        raise Exception("Curves don't have a similar number of points.")
    
    xy1 = Curve1.xy
    xy2 = Curve2.xy
    
    xy1 = resample(xy1, Nsample)
    xy2 = resample(xy2, Nsample)
    
    
    diff = sampleFunction(xy1, xy2)

    return diff

def compareHys(Hys1:Hysteresis, Hys2:Hysteresis):
    """
    

    Parameters
    ----------
    Hys1 : Hysteresis Object
        The first Hysteresis object.
    Hys2 : Hysteresis Object
        The second Hysteresis object.
    combineDiff : function, optional
        The function used to combine the differences for each cycle into a 
        single value. The default is defaultCombineDiff, which is the average
        distance between each point.


    Returns
    -------
    netdiff : float
        The average difference between both curves for the entire object.
    CycleDiffs : array
        The average difference between both curves for each cycle.

    """
    
    combineDiff = env.environment.fCombineDiff


    if Hys1.NCycles != Hys2.NCycles:
        raise Exception("Hysteresis don't have a similar number of Cycles.")    
    # Check both hystesis have the same number of reversals
    
    Cycles1 = Hys1.cycles
    Cycles2 = Hys2.cycles
    
    NCycles = Hys1.NCycles
    
    cycleDiffs = [None]*NCycles
    for ii in range(NCycles):
        cycleDiffs[ii]= compareCycle(Cycles1[ii], Cycles2[ii])
        
    netdiff = combineDiff(cycleDiffs)

    return netdiff, cycleDiffs
    






import numpy as np

from .curve import SimpleCurve
from .baseFuncs import concatenateHys


def getReturnCycle(cycleStart, cycleReturn):
    """
    This function finds the return cycle that closes a hystresis Full cycle

    """
    
    xy1 = cycleStart.xy
    # Get the and the max value
    x1 = cycleStart.xy
    x1max = x1[0,0]
    
    xy2 = cycleReturn.xy
    x2 = xy2[:,0]
    TransitonIndex = np.argmax(x2 < x1max)
    TransitonIndex = np.argmin(x2 < x1max)
    
    xyOut = np.zeros([TransitonIndex + 1, 2])
    xyOut[:TransitonIndex,:] = xy2[:TransitonIndex,:]
    xyOut[-1,:] = xy1[0,:]
    
    return SimpleCurve(xyOut)




def exandHysTrace(hysteresis, loadProtocolNcycles, skipStart = 1, 
                  skipEnd = 1):
    """
    
    The first cycle is assumed to be going from left to right (->).
    
    
    This function expands the trace of a hysteresis, where the trace has only
    one cycle per displacement in the load protocol.
    For example, if for a cycle N = 2, we go from:
        | . <-- .
        | . -----> .
        | . <----- .
        |     .--> .
    To:
        |     .<-- .
        | . -----> .
        | . <----- .
        | . -----> .
        | . <----- .
        |     .--> .
       
    We assume that the trace has a start cycle, and a end cycle. These start
    and end cycles are not expanded. The final/failure cycle can occur in the 
    expansion or return direction. To ensure they are not expanded, the user
    must skip the final cycles.
    
    Where cycles don't form a closed loop, the loop will be artificially closed
    by finding the closet point to the end of the cycle
   
    
    Parameters
    ----------
    hysteresis : Hystresis
        The input hysteresis to be expanded.
    loadProtocolNcycles : list
        The number of cycles for each reversal point. Has N/2 - 1 values in it.
        Each value of the list is the number of repeats in the output hysteresis.
        [n1, n2, n3, ... , n4]
    skipStart : int, optional
        Skip this many cycles at the start. The default is 0, which skips no steps.
    skipEnd : int, optional
        Skip this many cycles at the end. The default is 1, which skips no steps.
    FinalCyclePos : Bool, optional
        If the final cycle (the failure cycle) doesn't have a revesal point,
        Then this is set to true. The default is True.


    Returns
    -------
    Hysteresis
        The output hysteresis with the expanded load protocol.

    """
    
    # !!!: the final cycle is added back if we skip failure!

    # Get the cycles to be expanded
    cycles = hysteresis.cycles
    
    # Skip the requested start and end cycles.
    toExpand = cycles[skipStart:]
    if skipEnd != 0 :
        toExpand = toExpand[:-skipEnd]
    
    # Find the number of cycles that can be expanded
    NExpand = len(toExpand)
    NExpandIn = len(loadProtocolNcycles)
    
    Check = NExpand - 2*NExpandIn
    
    if Check != 0:
        print(NExpand,  2*NExpandIn)
        raise Exception('The number of entries in the load protocl does not match the '
                        'number half Cycles cycles to be expanded.')
        
        
    # Get the lenth of the output array
    xyList = [None]*int(np.sum(loadProtocolNcycles*2) + skipStart + skipEnd)


    # Create the output list of cycles to concatenate to a new Hysterersis.
    # nn corresponds to the original list, mm corresponds to the output list
    nn = 0
    mm = 0
    # The first cycles are unchanged
    xyList[:skipStart] = cycles[:skipStart]
    
    nn += skipStart
    mm += skipStart
    # All other cycles are expanded
    for Ncycles in loadProtocolNcycles:
        # Get the negative, positive, and positive exit cycles
        CycleNeg = cycles[nn]
        CyclePosExit = cycles[nn + 1]
        CyclePos = getReturnCycle(CycleNeg, CyclePosExit)

        # get the secondary cycles
        for ii in range(Ncycles):
            # The first cycle is always a Negative standard cycle
            xyList[mm] = CycleNeg
            
            # The second cycle my be a exit cycle. It it isn't, it's a truncated
            # positive cycle
            if ii != Ncycles - 1:
                xyList[mm + 1] = CyclePos
            else:
                xyList[mm + 1] = CyclePosExit
            mm += 2
        
        nn += 2
    
    if skipEnd != 0 :
        xyList[-skipEnd:] = cycles[-skipEnd:]

    return concatenateHys(xyList)

def _checkNcycles(MonotonicProtocol, loadProtocolNcycles):
    
    if isinstance(loadProtocolNcycles, int):
        return np.ones_like(MonotonicProtocol, int) * loadProtocolNcycles
    
    elif isinstance(loadProtocolNcycles, list) or isinstance(loadProtocolNcycles, np.ndarray):
        return np.array(loadProtocolNcycles, int)
        
    else:
        raise Exception('The input loadProtocolNcycles must be a integer, list, or numpy array')

def _getCyclePoints(P1, P2, Nsteps):
    np.linspace(P1, P1)


def createProtocol(MonotonicProtocol, loadProtocolNcycles, halfCycle = False):
    """
    Creates a reverse cyclic load protocol using a monotonic load protocol.
    In this context, the protocol is just the peak
    

    Parameters
    ----------
    MonotonicProtocol : list
        The load protocol amplitude values.
    loadProtocolNcycles : list, or int
        The number of cycles for each reversal point.
        if an integer all cycles will be expanded by the input integer.
        
        If it's a list, each monotonic cycle 'x' will be expanded by Nx
        [N1, N2, N3, ... , N4]
        
    halfCycle : bool
        A toggle that changes where the cycle returns to. If set to True, the load protocol will 
        return to 0 instead of the negative of the start value.

    Returns
    -------
    outputProtocol : array
        A np array of the input load protocol.

    """
    
    """
    Nsteps : int, optional
        The number of steps in the between each input point. Must be two or greater.
        The default is 2.    
    """
    
    # Check the number of expansions to use per cycle
    loadProtocolNcycles = _checkNcycles(MonotonicProtocol, loadProtocolNcycles)
    
    Ncycle = np.sum(loadProtocolNcycles)
    
    if len(MonotonicProtocol) != len(loadProtocolNcycles):
        raise Exception("The number of cycles isn't specified for each cycle in the monotonic load Protocol.")
        
    outputProtocol = np.zeros(Ncycle*2 + 1)
    nn = 1
    negPeak = 0
    for ii in range(len(MonotonicProtocol)):
        posPeak = MonotonicProtocol[ii]
        if not halfCycle:
            negPeak = -posPeak           
        
        for jj in range(loadProtocolNcycles[ii]):
            outputProtocol[nn] = posPeak
            outputProtocol[nn + 1] = negPeak
            nn += 2

    return outputProtocol





import numpy as np
from scipy.interpolate import interp1d
from .curve import Hysteresis, SimpleCurve, MonotonicCurve



# This is kind of unorganized right now...


# Todo:
    # make a concatenate to Cycle object.

def concatenate(curves, outputClass = Hysteresis):
    """
    This function creates a new curve from the xy data of a series of 
    curves, or xy arrays.
    
    The output object is new - if curves are used in the concatenation, 
    propreties from those curves, i.e. peaks, revesral points, etc, will 
    not be included in the output curve.
    
    If the last point of one curve is the first point of the next point, 
    that point will be skipped.

    Parameters
    ----------
    curves : list of curves
        A number of curves, or numpy arrays. Curves must a have a xy attribute
        numpy arrays must be 2D

    Returns
    -------
    hysteresis : Hysteresis Object
        The ouput hysteresis object which is a combination of all of the input
        curves

    """  
    
    
    xyList = [None]*len(curves)   
    for ii, vector in enumerate(curves):
        # Try to read the xy data from a monotonic curve object
        try:
            tempxy = vector.xy
        except:
            tempxy = vector
            
        # for curves after the first curve, we skip the first value
        if ii >= 1 and np.all(xyList[ii-1][-1] - vector[0] == 0):
            xyList[ii] = tempxy[1:,:]
        else:
            xyList[ii] = tempxy
    
    # Create new hysteresis, then add objects to the list    
    xy = np.concatenate(xyList)
    output = outputClass(xy)
    return output


def concatenateHys(curves):
    """
    This function creates a new hysteresis from the xy data of a series of 
    monotonic curves, or xy curves.
    
    If curves are used in the concatenation, propreties from those curves 
    will be lost.

    Parameters
    ----------
    curves : a list of SimpleCycle objects, or XY data
        A number of curve objects to be combined into a hysteresis.
        These curves either be 2D list/arrays, or objects with a xy attribute.

    Returns
    -------
    hysteresis : Hysteresis Object
        The ouput hysteresis object which is a combination of all of the input
        curves

    """
    return concatenate(curves)


# =============================================================================
# resample functions
# =============================================================================

def _linInterp(x,y, Nsamples):
    """
    A linear interpolation function that takes a target input curve to a 
    target sample curve. The sample curve goes between x0 and xN with Nsamples.
    """
    
    f = interp1d(x, y)
    
    outputx = np.linspace(x[0],x[-1], Nsamples)
    outputy = f(outputx)
    outputxy = np.column_stack((outputx, outputy))    
    return outputxy

def _linInterpSample(x,y, xSamples):
    """
    A linear interpolation function that takes a target input curve to a 
    target sample curve. Output of the curve are given at each sample point 
    """
    
    f = interp1d(x, y)
    
    outputy = f(xSamples)
    outputxy = np.column_stack((xSamples, outputy))    
    return outputxy


def _getNsamples(Targetdx, dxNet):
    if Targetdx >= abs(dxNet/2):
        print('Targetdx is larger than dxNet/2, no intermediate points made for the target dx = ' +str(dxNet))
        Nsamples = 2
    else:
        Nsamples = int(round(abs(dxNet/Targetdx))) + 1
    return Nsamples



# =============================================================================
# Tools for creating a load protocol
# =============================================================================

def _RemoveNeg(x, y, direction):
    
    difference = np.append(0, np.diff(x))
    
    condition = np.where(0 <= difference*direction)
    
    xOut = x[condition]
    yOut = y[condition]
    
    xy = np.column_stack([xOut, yOut])
    
    return xy

def removeNegative(Curve):
    """
    Removed values where the curve moves "backwards".

    Parameters
    ----------
    Curve : Hystersis Curve or array
        The input curve to remove negative values for.

    Returns
    -------
    Output : TYPE
        DESCRIPTION.

    """

    
    
    # Get sample parameters, then pass those to the new curve.
    
    # if the curve is a SimpleCycle
    if isinstance(Curve, SimpleCurve):
    
        x = Curve.xy[:,0]
        y = Curve.xy[:,1]
        direction = Curve.direction
        Output = SimpleCurve(_RemoveNeg(x, y, direction))    
       
    # if the curve is a hysteresis, we recursively create a series of Cycles
    elif isinstance(Curve, Hysteresis):
        outputCycles = [None]*Curve.NCycles
        for ii, Cycle in enumerate(Curve.Cycles):
            outputCycles[ii] = removeNegative(Cycle)
        Output = concatenateHys(*outputCycles)    # If Curve

    # if the curve is a Monotonic Cycle
    elif isinstance(Curve, MonotonicCurve):
    
        x = Curve.xy[:,0]
        y = Curve.xy[:,1]
        direction = Curve.direction
        
        print('Monotonic curves should have no reversals...')
        Output = MonotonicCurve(_RemoveNeg(x, y, direction))  
    
    # if it is a np array
    elif isinstance(Curve, np.ndarray):
        x = Curve[:,0]
        y = Curve[:,1]
        
        # TODO: Create a standardized get direction function
        if x[0] <= x[-1]:
            direction = 1
        else:
            direction = -1
        
        Output = _RemoveNeg(x, y, direction)

    return  Output




# def get




import hysteresis.baseFuncs as hys
from hysteresis.curve import Hysteresis, SimpleCurve
import numpy as np




x = np.linspace(0,1,101)
xy1 = np.column_stack([x,x])
xy2 = np.column_stack([x+1, x[::-1]])
# xy2 = np.column_stack([x+1, x[::-1]])
outputCurve1 = np.concatenate((xy1[:-1],xy2))
outputCurve_noskip = np.concatenate((xy1,xy1))


def test_concatenateHys():
    out = hys.concatenateHys((xy1,xy2))
    test1 = np.sum(out.xy - outputCurve1) < 0.0001
    test2 = isinstance(out, Hysteresis)
    assert test1 and test2

def test_concatenate():
    out = hys.concatenate((xy1, xy2), outputClass=SimpleCurve)
    print(out.xy)
    test1 = np.sum(out.xy - outputCurve1) < 0.0001
    test2 = isinstance(out, SimpleCurve)
    assert test1 and test2
    
def test_concatenate_noSkip():
    out = hys.concatenate((xy1, xy1), outputClass=SimpleCurve)
    test1 = np.sum(out.xy - outputCurve_noskip) < 0.0001
    test2 = isinstance(out, SimpleCurve)
    assert test1 and test2    
    




# -*- coding: utf-8 -*-
"""
Created on Sat Aug 31 15:13:13 2019
@author: Christian


"""
import numpy as np
import hysteresis as hys
from hysteresis.envelope import _LPparser
from scipy.signal import savgol_filter
import pytest

# =============================================================================
# Load initial stuff
# =============================================================================

data = np.loadtxt('Trial2.csv', skiprows = 2, delimiter  =',')
x = data[:,0]
x = savgol_filter(x,15,1)
y = data[:,1]
xy = np.column_stack([x,y])
lp = [5,3,3,3,3,3,3,3, 1]
myHys = hys.Hysteresis(xy)


Lp2 = [6,6,6,4,2,2,2,2,2]

# =============================================================================
# Find the backbone
# =============================================================================

def test_lpsteps_Nstep():
    """
    This test checsk if the load protocal adds indexieies as expected.
    Results are only returned short of the final cycle in the experiment, which
    ends short.
    """
    parse = _LPparser(lp)
    solution = np.array([0,1,6,9,12,15,18,21,24, 27])
    
    diff = np.sum(np.abs(solution - parse))
    
    # Add one because the first index is included
    assert len(parse) == (len(lp) + 1) 
    assert diff < 10**-6


def test_lpsteps_Nstep2():
    parse = _LPparser(Lp2)
    solution = np.array([0,1,7,13,19,23,25,27,29,31])
    
    diff = np.sum(np.abs(solution- parse))
    
    assert len(parse) == (len(Lp2)+1)
    assert diff < 10**-6


def test_inputParse():
    """makes sure we get an error if the user doesn't ask for one of end points
    or peaks."""
    # Function: get envelope curve
    analHys = hys.Hysteresis(xy)
    # analHys.plotVsIndex(True)
    
    with pytest.raises(ValueError) as errorInfo:
    
        # Get the backbone
        backbone = hys.getBackboneCurve(analHys, lp, False, False)
        # assert True

    assert errorInfo.type is ValueError


def test_backbone():
    # Function: get envelope curve
    analHys = hys.Hysteresis(xy)
    
    # Get the backbone
    hys.getBackboneCurve(analHys, lp, True)
    hys.getBackboneCurve(analHys, lp, True,includeNegative=True)

    assert True == True

# =============================================================================
# Fit the backbone curve.
# =============================================================================

def test_fit_avg():
    
    avg, pos, neg = hys.getAvgBackbone(myHys, lp, True, True)   
    avg.setArea()
    avg.plot()
    A1 = avg.getNetArea()
    
    Curve = hys.fitEEEP(avg)
    Curve.setArea()
    A2 = Curve.getNetArea()
    
    check1 = abs(A1 - 6622.37) < 0.01
    check2 = abs(A2 - 5775.33) < 0.01
    # check2 = (A1 - A2) < 10**-6
    assert np.all((check1, check2))



def test_negative_BB():
    backbone = hys.getBackboneCurve(myHys, lp,1,1, includeNegative=True)
    
    xyReversal = myHys.getReversalxy()
    
    finalPos = xyReversal[-3]
    finalNeg = xyReversal[-2]
    
    xy = backbone.xy
    
    posDiff = sum(np.abs(finalPos - xy[-1]))
    negDiff = sum(np.abs(finalNeg - xy[0]))
    
    assert posDiff < 0.01 and negDiff < 0.01


def ManualTest():
    """
    lets us visually inspect if anything seems wrong
    """
    avg, pos, neg = hys.getAvgBackbone(myHys, lp)   
    backbone = hys.getBackboneCurve(myHys, lp,1,1, includeNegative=True)

    
    fig, ax = myHys.initFig()
    myHys.plot()
    backbone.plot()
    # neg.plot()

if __name__ == '__main__':
    ManualTest()

    test_lpsteps_Nstep()
    test_lpsteps_Nstep2()
    test_inputParse()
    test_backbone()
    test_fit_avg()
    test_negative_BB()





"""
Test if intersections can be found.
"""

import hysteresis.curve as hys
import numpy as np
from pytest import approx

# out = test_Curve3_Area()


def makeCurve(f):
    "lamda function that creates all curves"
    xdata = np.linspace(0,4,1001)
    y = f(xdata)
    xy = np.column_stack([xdata, y])
    Curve = hys.Curve(xy)   
    return Curve    

def function1(x):
    return x - 2

def function2(x):
    return x**3 - 5*x**2 + 5*x + -1

def function3(x):
    return 3 + (np.e)**x  
    


def test_data_Curve1():
    curve = makeCurve(function1)
    interInds = hys.data.getIntersections(curve.y)


    xyinter = curve.getXIntersections()
    assert len(interInds) ==1
    
def test_Curve1():
    curve = makeCurve(function1)

    xyinter = curve.getXIntersections()
    assert len(xyinter) ==1
    assert 2 == approx(xyinter[0][0],0.01)
    assert 0 == approx(xyinter[0][1],0.01,abs=True)
    
def test_data_Curve2():
    curve = makeCurve(function2)
    xyinter = hys.data.getIntersections(curve.y)
    
    assert len(xyinter) == 3

def test_Curve2():
    curve = makeCurve(function2)
    xyinter = curve.getXIntersections()
    
    assert len(xyinter) == 3    
    # assert 2 == approx(xyinter[0][0],0.01)
    assert 0 == approx(xyinter[0][1],0.01,abs=True)
    assert 0 == approx(xyinter[1][1],0.01,abs=True)
    assert 0 == approx(xyinter[2][1],0.01,abs=True)
    
    
    
def test_data_Curve3():
    curve = makeCurve(function3)
    xyinter = hys.data.getIntersections(curve.y)
    
    assert len(xyinter) == 0
   
def test_Curve3():
    curve = makeCurve(function3)
    xyinter = curve.getXIntersections()
    assert len(xyinter) == 0



def test_BMD_Curve():
    xyBmd = np.loadtxt('bmd.csv', delimiter=',')
    extraPoints = [[10,0],[10,0]]
    
    xyBmd = np.concatenate((xyBmd, extraPoints))
    myHys = hys.Curve(xyBmd) 

    xyinter = myHys.getXIntersections()
    
    # myHys.plot()
    # import matplotlib.pyplot as plt
    # plt.plot(xyinter[:,0], xyinter[:,1])
    assert len(xyinter) == 3


if __name__ == "__main__":
    # test_data_Curve1()
    # test_Curve1()
    
    # test_data_Curve2()
    # test_Curve2()
    
    # test_data_Curve3()
    # test_Curve3()
    
    test_BMD_Curve()


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt
import scipy
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
# Note, 2D array


data =  np.array([[-0.724, -0.834], [-0.680, -1.239], [-0.479, -1.032],
                  [-0.488, -0.871], [-0.526, -0.774], [-0.331, -0.741], 
                  [-0.008, -0.596], [0.060, -0.430],  [0.204, -0.256],
                  [0.289, 0.024],   [0.369, 0.234],   [0.554, 0.655],
                  [0.545, 0.724],   [0.635, 0.857],   [0.564, 0.868],
                  [0.368, 0.858],   [0.385, 0.725],   [0.349, 0.636],
                  [0.181, 0.496],   [0.042, 0.399],   [-0.204, 0.120],
                  [-0.380, 0.002],  [-0.429, -0.131], [-0.655, -0.421]])

Npoints = len(data)
t = np.arange(0, Npoints)/Npoints

climateHys = hys.SeasonalCurve(data, 12, 1,0)

fig, ax = climateHys.initFig()
climateHys.plotScatter()
climateHys.plotFittedCurve()

fig, ax = climateHys.initFig()
climateHys.plotTemp()
climateHys.plotDTemp()
# Curve.plotFittedCurve()

Tcoefs = climateHys.tempCoef
dTcoefs = climateHys.dTempCoef

climateHys.printT()
climateHys.printDT()

# def f(x, a, b):
#     return a*x + b

# def g(t, b, c, Cx):
#     return b*np.cos(2*np.pi*t + c) + Cx

# def geth(n = 1, m = 1):

#     def h(t, a, b,c,d, Cy):
#         return a*np.cos(2*np.pi*t + b)**n + c*np.sin(2*np.pi*t + d)**m + Cy
    
#     return(h)

# h = geth()

# out = curve_fit(f, data[:,0], data[:,1])
# out2 = curve_fit(g, t, data[:,0])
# out3 = curve_fit(h, t, data[:,1])

# fig, ax = plt.subplots()
# # plt.plot(data[:,0], data[:,1], '.', linewidth = 0)
# # plt.plot(data[:,0], f(data[:,0], *out[0]))
# # plt.plot(g(t, *out2[0]), data[:,1])
# plt.plot(t, g(t, *out2[0]))
# plt.plot(t, h(t, *out3[0]))

# fig, ax = plt.subplots()
# plt.plot(data[:,0], data[:,1], '.', linewidth = 0)
# plt.plot(g(t, *out2[0]), h(t, *out3[0]))
# # smallTriangles.setArea()
# # Anet = smallTriangles.getNetArea()
# # 0.004982994450842999


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import numpy as np
import hysteresis as hys
import matplotlib.pyplot as plt


t  = np.linspace(0,4,1000)*np.pi
x  = np.sin(t)
y  = np.cos(t)*t
xy = np.column_stack([x,y])

myHys = hys.Hysteresis(xy)

# =============================================================================
# 
# =============================================================================


def test_cycles_plot(monkeypatch):
    """ Tests if the cycles can plot correctly"""
    monkeypatch.setattr(plt, 'show', lambda: None)
    myHys.cycles[0].setSubCycles()
    myHys.cycles[0].plot(showPeaks=True, showReversals=True)
    assert True == True

def test_subcycles_plot(monkeypatch):
    """ Tests if the cycles can plot correctly"""
    
    monkeypatch.setattr(plt, 'show', lambda: None)
    myHys.cycles[0].plotSubCycles(showPeaks=True)
    
    assert True == True




import numpy as np


from .curve import Curve

import matplotlib.pyplot as plt
from scipy.optimize import curve_fit



# =============================================================================
# Curve objects
# =============================================================================



# T is normalized to 1/T

def getTemp(t, a, b, Cx):
    return a*np.cos(2*np.pi*t + b) + Cx

def getDeltaTemp(n = 1, m = 1):

    def h(t, a, b, c, d, Cy):
        return a*np.cos(2*np.pi*t + b)**n + c*np.sin(2*np.pi*t + d)**m + Cy
    
    return(h)

def defaultFitCurve(f, xy, **kwargs):
    """
    A simple curve fiting implementation using Scipy's nonlinear least 
    squares method.

    Parameters
    ----------
    f : function
        The function to be fit.
    xy : array
        The 2D input array of x/y data.
    **kwargs : kwargs
        The keyword arguements to be passed to curve_fit.

    Returns
    -------
    coef : TYPE
        The output coefficients for the fit function.
    correlation : TYPE
        The output correlation matrix for the fit variables.

    """
    
    coef, correlation = curve_fit(f, xy[:,0], xy[:,1], **kwargs)
    
    return coef, correlation


class SeasonalCurve(Curve):
    """
    The input XY data is assumed to be evenly spaced in time.
    
    The curve will be closed so that the first point is equal to the final
    point.
    """
    
    
    def __init__(self, xy, period = 1, n = 1, m = 1):
        Curve.__init__(self, xy)
        
        self.time = np.arange(0, self.Npoints)/self.Npoints
        self.period = period
        self.n = n
        self.m = m
        
        self.tempFunction = getTemp
        self.dTempFunction = getDeltaTemp(n,m)
        
        self.tempCoef, self.tempCoer = curve_fit(self.tempFunction, self.time, xy[:,0])
        self.dTempCoef, self.dtempCoer  = curve_fit(self.dTempFunction, self.time, xy[:,1])
        
        
        self.fittedT = self.tempFunction(self.time, *self.tempCoef)
        self.fittedDT= self.dTempFunction(self.time, *self.dTempCoef)
        
        # Close the curve
        self.time = np.append(self.time, 1)
        self.fittedT = np.append(self.fittedT,self.fittedT[0])
        self.fittedDT = np.append(self.fittedDT,self.fittedDT[0])
        
        
        self.Cx = self.tempCoef[-1]
        self.Cy = self.dTempCoef[-1]
        

    def plotScatter(self):
        x = self.xy[:,0]
        y = self.xy[:,1]
        plt.plot(x, y, '.', linewidth = 0)

    def plotFittedCurve(self, xlim = [], ylim = []):
        
        x = self.fittedT
        y = self.fittedDT
        plt.plot(x, y)
        
        # return self.plotfunction(self, x ,y, 0, 0, xlim, ylim, 0)
    
    def plotTemp(self, xlim = [], ylim = []):
        
        
        x = self.time*self.period
        y = self.fittedT
        plt.plot(x, y)
        # return self.plotfunction(self, x ,y, 0, 0, xlim, ylim, 0) 
    
    
    def plotDTemp(self, xlim = [], ylim = []):
        
        x = self.time*self.period
        y = self.fittedDT
        plt.plot(x, y)
        # return self.plotfunction(self, x ,y, 0, 0, xlim, ylim, 0)     
    
    def printT(self):
        print('The expression for temperature is: ')
        print('a * np.cos(2*np.pi*t / T + b) + Cx')
        
        print('Where the coeficients are:')
        print('a = ', self.tempCoef[0])
        print('b = ', self.tempCoef[1])
        print('Cx = ', self.tempCoef[2])

    
    def printDT(self):
        
        print('The expression for delta temperature is: ')
        print('a * np.cos(2*np.pi*t / T + b)**n + c*np.sin(2*np.pi*t / T + d)**m + Cy')
        
        print('Where the coeficients are:')
        print('n = ', self.n)
        print('m = ', self.m)        
        print('a = ', self.dTempCoef[0])
        print('b = ', self.dTempCoef[1])
        print('c = ', self.dTempCoef[2])
        print('d = ', self.dTempCoef[3])
        print('Cy = ', self.dTempCoef[4])    
    
    



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt
        
disp = np.loadtxt('UFP_Disp.out',delimiter=' ')
force = np.loadtxt('UFP_RFrc.out',delimiter=' ')
testHys2 = np.column_stack([disp[:,1], -force[:,1]])

def test_DownSampled_Plotting(monkeypatch):
    """ Tests if the resampled hystresis can plot correctly """
    
    monkeypatch.setattr(plt, 'show', lambda: None)
    DamperHys = hys.Hysteresis(testHys2)
    
    downsampledHys = hys.resample(DamperHys, 20)
    downsampledHys.plot(showReversals= True)
    plt.close()
    downsampledHys.plotCycles([2,3])
    plt.close()
    downsampledHys.plotArea()
    plt.close()
    downsampledHys.plotSlope()    
    plt.close()
    
    assert True == True


# test_DownSampled_Plotting()


# -*- coding: utf-8 -*-
"""
Created on Sat May  1 23:49:09 2021

@author: Christian
"""


import numpy as np
import hysteresis.plotSpecial.animate as ani
import hysteresis as hys
import pytest

np.random.seed(101)
x = np.linspace(0, 1, 101)*10
y1 = np.sin(x)
y2 = np.cos(x)

x2 = np.linspace(0, 1, 10)

trianglexy1 = np.column_stack((x, y1))
trianglexy2 = np.column_stack((x, y2))
trianglexy3 = np.column_stack([x2, x2])
test1 = hys.Hysteresis(trianglexy1)
test2 = hys.Hysteresis(trianglexy2)
test3 = hys.Hysteresis(trianglexy3)


def test_validate_pass():
    
    curves = [test1, test2]
    ani.JointAnimation(curves)


def test_validate_fail():

    with pytest.raises(Exception):
        curves = [test1, test2, test3]
        ani.JointAnimation(curves)




import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt

def makeCurve():
    "lamda function that creates all curves"
    t = np.linspace(0,4,1000)*np.pi
    x = np.cos(t)
    y = np.sin(t)*t
    
    xy = np.column_stack([x,y])
    Curve = hys.Hysteresis(xy)
    return Curve    



def test_expandHysTrace():
    curve = makeCurve()
    expand = hys.exandHysTrace(curve, [2])
    
    assert len(expand.cycles) == 6





# curve = makeCurve()
# curve.plotVsIndex()
# expand = hys.exandHysTrace(curve, [2])
# expand.plotVsIndex()
# test_expandHysTrace()

# curve = makeCurve()
# curve.plot()
# curve.plotLoadProtocol()
# fig, ax = curve.plotCycles()
# lines = fig.axes[0].lines
# for line in lines:
#     plt.setp(line, linestyle='--')

# print(curve.NCycles)
# # print(np.round(curve.loadProtocol))



# expand = hys.exandHysTrace(curve, [2])
# expand.plot()
# expand.plotLoadProtocol()
# fig, ax = curve.plotCycles()
# lines = fig.axes[0].lines
# for line in lines:
#     plt.setp(line, linestyle='--')
    






import numpy as np
from hysteresis import data
import hysteresis.env as env
import matplotlib.pyplot as plt


class CurvePlotter:
    """
    The Curve plotter class is used to define plotting behaviour that is common 
    to all types of curves. It is generally ment to be inherited from and not
    used directly. 
    """
    
    
    def plot(self, showReversals = False, showPeaks = False, labelReversals = None,
             **kwargs):
        """
        Used the plot function to make a xy plot of the curve. The default 
        plot function uses matplotlib for plotting, and returns the plotted
        line.
        
        Parameters
        ----------
        showReversals : bool
            A switch that specifics if cycle reversal points should be plotted.
        showPeaks : bool
            A switch that specifics if the detected peak values should be plotted.
        labelReversals : list or 'all', optional
            A list of the cycles to be labled. 
            A value of 'all' can also be specified, in which case all cucles will be plotted.
            The default is [], which labels no cycles    
        kwargs : list, optional
            Any additional keyword arguements will be passed to the matplotlib
            plot function. This can be used to custize char appearance.            

        Outputs
        -------
        Matplotlib Line

        """
        
        if labelReversals is None:
            labelReversals = []

        return self.plotFunction(self.xy, showReversals, showPeaks, self.peakIndexes, 
                                 self.reversalIndexes, labelReversals,
                                 **kwargs)
                
    def plotVsIndex(self, showReversals = False, showPeaks = False, 
                     labelReversals = None, **kwargs):
        """
        Used the plot function to make a plot of each x point vs. it's index.
        This can be useful for understanding how x values change over time.
        The default plot function uses matplotlib for plotting, and returns 
        the plotted line.
        
        Parameters
        ----------
        showReversals : bool
            A switch that specifics if cycle reversal points should be plotted.
        showPeaks : bool
            A switch that specifics if the detected peak values should be plotted.
        labelReversals : list or 'all', optional
            A list of the cycles to be labled. 
            A value of 'all' can also be specified, in which case all cucles will be plotted.
            The default is [], which labels no cycles     
        kwargs : list, optional
            Any additional keyword arguements will be passed to the matplotlib
            plot function. This can be used to custize char appearance.
            
        """           
        
        if labelReversals is None:
            labelReversals = []
                
        x = np.arange(0,len(self.xy[:,0]))
        y = self.xy[:,0]
        
        xy = np.column_stack([x,y])
                    
        return self.plotFunction(xy, showReversals, showPeaks, self.peakIndexes, 
                                 self.reversalIndexes, labelReversals, **kwargs)

    def plotLoadProtocol(self, comparisonProtocol = [], **kwargs):
        """
        Plots the peak x values for each cycle in a curve.
        An additional load protocol can be provided, so the user can comapare the
        curves peak x points against an expected load protocol.
        
        The default plot function uses matplotlib for plotting, and returns 
        the plotted line.
        
        Parameters
        ----------
        comparisonProtocol : array
            A load protocol that will be plotted on the same figure. Useful
            for comparing the curves load protcol to the input protocol.
        kwargs : list, optional
            Any additional keyword arguements will be passed to the matplotlib
            plot function. This can be used to custize char appearance.
        """           
        showReversals = False
        showPeaks = False
        labelCycles = []
        y = self.loadProtocol
        x = np.arange(0,len(y))
        xy = np.column_stack([x,y])
        
        line = self.plotFunction(xy, showReversals, showPeaks, labelCycles, **kwargs)
        
        if len(comparisonProtocol) != 0:
            line2 = plt.plot(comparisonProtocol)
            return [line, line2]

        return line
    
    def plotSlope(self,  showReversals = False, showPeaks = False, 
                  labelReversals = None, **kwargs):
        """
        Used the plot function to make a xy plot of the slope at each point of 
        the curve. The slope is calcualted using the slope function, which
        uses a centeral finite difference scheme by default.
        plot function uses matplotlib for plotting, and returns the plotted
        line.
        
        Parameters
        ----------
        showReversals : bool
            A switch that specifics if cycle reversal points should be plotted.
        showPeaks : bool
            A switch that specifics if the detected peak values should be plotted.

        labelReversals : list or 'all', optional
            A list of the cycles to be labled. 
            A value of 'all' can also be specified, in which case all cucles will be plotted.
            The default is [], which labels no cycles  
        kwargs : list, optional
            Any additional keyword arguements will be passed to the matplotlib
            plot function. This can be used to custize char appearance.
        """          
        
        xy = np.column_stack([self.xy[:,0], self.slope])
        
        if labelReversals is None:
            labelReversals = []
    
        return self.plotFunction(xy, showReversals, showPeaks, self.peakIndexes, 
                                 self.reversalIndexes, labelReversals,
                                 **kwargs)    
    
    def plotArea(self,  showReversals = False, showPeaks = False, 
                 labelReversals = None, **kwargs):
        """
        Used the plot function to make a xy plot of the area under each point of 
        the curve. The area is calculated using the area function, which
        uses the midpoint rule by default.
        The plot function uses matplotlib for plotting, and returns the plotted
        line.
        
        Parameters
        ----------
        showReversals : bool
            A switch that specifics if cycle reversal points should be plotted.
        showPeaks : bool
            A switch that specifics if the detected peak values should be plotted.

        labelReversals : list or 'all', optional
            A list of the cycles to be labled. 
            A value of 'all' can also be specified, in which case all cucles 
            will be plotted.
            The default is [], which labels no cycles
        kwargs : list, optional
            Any additional keyword arguements will be passed to the matplotlib
            plot function. This can be used to custize char appearance.            
            
        """       
        
        if labelReversals is None:
            labelReversals = []
            
        xy = np.column_stack([self.xy[:,0], self.area])

        return self.plotFunction(xy, showReversals, showPeaks, self.peakIndexes, 
                                 self.reversalIndexes, labelReversals,
                                 **kwargs)    
                        
    def plotCumArea(self,  showReversals = False, showPeaks = False, 
                    labelReversals = None, cumulativeDisp = True, **kwargs):
        """
        Used the plot function to make a xy plot of cumulative area at each
        point of the curve. The cumulative area is a summation of the areas up
        to the point in question.
        The area is calculated using the area function, which uses the midpoint 
        rule by default.
        The plot function uses matplotlib for plotting, and returns the plotted
        line.
        
        Parameters
        ----------
        showReversals : bool
            A switch that specifics if cycle reversal points should be plotted.
        showPeaks : bool
            A switch that specifics if the detected peak values should be plotted.
        labelReversals : list or 'all', optional
            A list of the cycles to be labled. 
            A value of 'all' can also be specified, in which case all cucles will be plotted.
            The default is [], which labels no cycles
        kwargs : list, optional
            Any additional keyword arguements will be passed to the matplotlib
            plot function. This can be used to custize char appearance.
        
        """   
        
        if labelReversals is None:
            labelReversals = []
            
        # We get the cumulative displacement and area
        if cumulativeDisp:
            x = self.getCumDisp()
        else:
            x = self.xy[:,0]
        y = self.getCumArea()
        xy = np.column_stack([x,y])

        return self.plotFunction(xy, showReversals, showPeaks, self.peakIndexes, 
                                 self.reversalIndexes, labelReversals,
                                 **kwargs)    
             
    def initFig(self, xlims = None, ylims = None):
        """       
        Initializes the plot using the default rules.
        """
        
        if xlims is None:
            xlims = []
        
        if ylims is None:
            ylims = []
            
        return self.initializeFig(xlims, ylims)


    def _plot_curves(self, curves, curveInds, colorCycles, **kwargs):
        """
        Used by plotCycles and plotSubCycles
        """
        Ncolor = len(colorCycles) 
        lines = []
        # If the list has content, get the cycles to plot, otherwise plot everything
        if curveInds is not None:
            curves = [curves[ii] for ii in curveInds]
            
        for ii, curve in enumerate(curves):
            c = colorCycles[int(ii % Ncolor)]
            line, = plt.plot(curve.xy[:,0], curve.xy[:,1], c=c, **kwargs)
            lines.append(line)
        return lines     

class CurveOperations:
    
    """
    The Curve operations class is used to define operations that all curves
    will have. This includes features such iteration, mathematical operations
    like addition or multiplaction, and setting/getting items.
    """
    
    __array_ufunc__ = None

    
    @property
    def y(self):
        return self.xy[:,1]
    
    @property
    def x(self):
        return self.xy[:,0]    
    
    def _parseXY(self, xy):
        if isinstance(xy, list):
            xy = np.column_stack(xy)
        return xy
    
    def __len__(self):
        return len(self.xy[:,0])

    def __iter__(self):
        return iter(self.xy)
    
    def __getitem__(self, ind):
        return self.xy[ind]
    
    def __setitem__(self, ind, val):
        self.xy[ind] = val

    def _getInstance(self):
        return type(self)
    
    def _convertToCurve(self, other):
        """
        Converts non-hysteresis datatypes 
        """
        if isinstance(other, np.ndarray):
            return Curve(other)

    def __add__(self, other):
        """ Enables addition of curve x values"""
        Instance = self._getInstance()
        operand = _getOperand(other)
        x = self.xy[:,0]
        y = self.xy[:,1]+operand        
        
        return Instance(np.column_stack([x,y]))
    
    def __sub__(self, other):
        Instance = self._getInstance()
        operand = _getOperand(other)
        x = self.xy[:,0]
        y = self.xy[:,1] - operand        
        
        return Instance(np.column_stack([x,y]))
    
    def __rsub__(self, other):
        Instance = self._getInstance()
        operand = _getOperand(other)
        x = self.xy[:,0]
        y = operand - self.xy[:,1]   
        
        return Instance(np.column_stack([x,y]))

    def __neg__(self):
        Instance = self._getInstance()
        x =  self.xy[:,0]
        y = -self.xy[:,1]           
        return Instance(np.column_stack([x,y]))
    
    def __mul__(self, other):
        """
        It would be useful of having a hystresis base state, then copying that over.
        """
        # Get the current instance of curve
        Instance = self._getInstance()
        operand = _getOperand(other)
        x = self.xy[:,0]
        y = self.xy[:,1]*operand
        
        return Instance(np.column_stack([x,y]))

    def __truediv__(self, other):
        # Get the current instance of curve
        Instance = self._getInstance()
        operand = _getOperand(other)
        x = self.xy[:,0]
        y = self.xy[:,1] / operand
        
        return Instance(np.column_stack([x,y]))

    def __rtruediv__(self, other):
        # Get the current instance of curve
        Instance = self._getInstance()
        operand = _getOperand(other)    
        x = self.xy[:,0]
        y = operand / self.xy[:,1]
        
        return Instance(np.column_stack([x,y]))

class Curve(CurveOperations, CurvePlotter):
    
    """
    The curve base object represents a generic xy curve with no limitations.
    
    All other curves inherit from the CurveBase class. It has access to methods
    all other curves should have access to. This uncludes plotting 
    functionality, slope functions, area functions, and displacement functions.
    
    Parameters
    ----------
    XYData : array
        The input array of XY date for the curve. A list [x,y] can also be 
        passed into the function.
    xunit : str, optional
        The units on the x axis. The default is ''.
    yunit : str, optional
        The units on the y axis. The default is ''.

    """
    
    colorCycles = ['C0', 'C0', 'C1', 'C1']
    reversalIndexes = np.array([0,-1])
    interIndexesX = None
    interIndexesY = None
    
    
    def __init__(self, XYData, xunit = '', yunit = ''):

        
        self.xy = self._parseXY(XYData)
        self.Npoints = len(self.xy[:,0])
        
        # The function to be used to calcualte area. These can be overwritten
        self.areaFunction = env.environment.fArea
        self.slopeFunction = env.environment.fslope
        self.lengthFunction = env.environment.flength
        self.plotFunction = env.environment.fplot
        
        self.initializeFig = env.environment.finit
        self.showCycles = env.environment.fcycles
        
        self.xunit = xunit
        self.yunit = yunit
        
        self.peakIndexes = None
        
    def setArea(self):
        """ sets the area under each point of the curve using the area function"""
        self.area = self.areaFunction(self.xy)
        return self.area
    
    def getCumDisp(self):
        """ 
        Gets the absolute value of cumulative displacement of the curve at each data point.
        """
        dx = np.diff(self.xy[:,0])
        return np.append(0, np.cumsum(np.abs(dx)))
    
    def getNetCumDisp(self, startIndex = 0, endIndex = 0):
        """ Gets the total cumulative displacement between the start and start and end 
        indexes. By default the whole curve us used.
        """
        dx = np.append(0, np.diff(self.xy[:,0]))
        
        if endIndex == 0:
            endIndex = self.Npoints
                   
        return np.sum(np.abs(dx[startIndex:endIndex]))
    
    def getCumArea(self):
        """ 
        Gets the cumulative area under the curve for the entire curve.
        """
        Area = self.area
        return np.cumsum(Area)
    
    def getNetArea(self, startIndex = 0, endIndex = 0):
        """
        Returns the net area between two indexes in the xy curve. The default 
        setting is to return the area for the whole cuve.

        """
        Area = self.area

        if endIndex == 0:
            endIndex = self.Npoints
                   
        return np.sum(Area[startIndex:endIndex])
           
    def setSlope(self):
        """
        Calcuates the slope of the curve at each point.
        The user can pass in a custom function that calculates the slope.
        
        By default, uses the basic slope funciton. Custom behaviour can be
        specified by modifying the environment slope function.        
        """
        
        # Calculate end point slope
        xy = self.xy
        self.slope  = self.slopeFunction(xy)

    def setLength(self):
        """
        Calcuates the length of the curve, that is, the distance from each 
        point to the next point. 
        Starts at point 0, and ther are there are N-1 length values for the N 
        data points.
        
        By default, uses the basic length funciton. Custom behaviour can be
        specified by modifying the environment length function.

        """
        
        # Calculate end point slope
        xy = self.xy
        self.length  = self.lengthFunction(xy)

    def setPeaks(self, peakDist = 2, peakWidth = None, peakProminence = None):
        """
        Finds the indexes of max and min points, then stores them.
        """
        
        y = self.xy[:,1]
        peakIndexes = data.getCycleIndexes(y, peakDist, peakWidth, peakProminence)        
        self.peakIndexes = peakIndexes
        
        xy = self.xy
        if xy[peakIndexes[0],1] < xy[peakIndexes[1],1]:
            self.minIndexes = peakIndexes[0::2]
            self.maxIndexes = peakIndexes[1::2]
        else:
            self.minIndexes = peakIndexes[1::2]
            self.maxIndexes = peakIndexes[0::2]
            
    def getPeakxy(self,):
        
        if self.peakIndexes is not None:
            return self.xy[self.peakIndexes]
        else:
            raise Exception('No peaks have been set')

            
    def setIntersectionInds(self):
        """
        Sets the intersection indicies in x and y.
        """
        self.interIndexesX =  data.getIntersections(self.x)
        self.interIndexesY =  data.getIntersections(self.y)
                    
    def getXIntersections(self, tol=1e-6):
        """
        Returns the points closest to the x axis intersections, i.e. where
        y = 0
        """
        if self.interIndexesY is None:
            interInds = data.getIntersections(self.y, tol)
        else:
            interInds = self.interIndexesY
        return self.xy[interInds]
        
            
    # def getYIntersections(self):
    #     """
    #     Returns the points closest to the y axis intersections, i.e. where
    #     x = 0
    #     """
    #     return data.getIntersections(self.y)
        
        
    
    
def _getOperand(curve):
    """
    Gets the operand (what data the function acts on) for operation functions
    (+,-,*,/)
    The data used in these functions depends on the input given.
    
    """
    if not hasattr(curve, '__len__'): # assume scalar if no length
        operand = curve
    elif hasattr(curve, 'xy'): # use the xy if it's a curve from the hysteresis module.
        operand = curve.xy[:,1]
    elif isinstance(curve, np.ndarray): # If a numpy array, do stuff depending on array dimension.
        if 1 == len(curve.shape):
            operand = curve
        elif 2 == len(curve.shape) and curve.shape[-1] == 2: # if 2D array
            operand = curve[:,1]
        else: # if 1D array
            raise Exception(f'{curve.shape[-1]}D curve give, only 1 or 2D supported')
    return operand

class Hysteresis(Curve):
    """
    Hysteresis objects are those that have at least one reversal point in 
    their data. Reversal points are those where the x axis changes direction.
    
    The hysteresis object has a number of functions that help find the reversal
    points in the x data of the curve.
    
    The hysteresis object also stores each each half-cycle (where there is no
    change in direction) as a SimpleCycle object.
    
    
    Parameters
    ----------
    XYData : array
        The input array of XY date for the curve. A list [x,y] can also be 
        passed into the function.
    revDist : int, optional
        Used to filter reversal points based on the minimal horizontal distance 
        (>= 1) between neighbouring peaks in the x axis. Smaller peaks are 
        removed first until the condition is fulfilled for all remaining peaks.
        The default is 2.
    revWidth : int, optional
        Used to filter reversal points using the approximate width in number of 
        samples of each peak at half it's prominence. Peaks that occur very 
        abruptly have a small width, while those that occur gradually have 
        a big width.
        The default is None, which results in no filtering.  
    revProminence : number, optional
        Used to filter reversal points that aren't sufficently high. Prominence 
        is the desired difference in height between peaks and their 
        neighbouring peaks. 
        The default is None, which results in no filtering.     
    setCycles : Boolean, optional
        Used to turn on or off setting the cycle revesal points. 
        Can be turned off for performance reasons.
    setArea : Boolean, optional
        Used to turn on or off setting the Area by default. 
        Can be turned off for performance reasons.
    setSlope : Boolean, optional
        Used to turn on or off setting the slopeby default. 
        Can be turned off for performance reasons.    
        
    """
    
    def __init__(self, XYData, revDist = 2, revWidth = None, revProminence = None,
                 setCycles = True, setArea = True, setSlope = True, **kwargs):
 
        Curve.__init__(self, XYData, **kwargs)
        # self.setReversalPropreties(revDist, revWidth, revProminence)
        self._setStatePropreties(revDist, revWidth, revProminence,
                                 setCycles, setArea, setSlope)
        
        #TODO Create warning if cycles don't make sense.
        self.cycles:list[SimpleCurve] = None
        if setCycles == True:
            self.setReversalIndexes(revDist, revWidth, revProminence)
            self.setCycles()
        
        if setArea ==True:
            self.setArea()
        if setSlope ==True:
            self.setSlope()
    
    def setReversalPropreties(self, revDist = 2, revWidth = None, 
                           revProminence = None):
        """
        Sets the propreties that are used to caculate where reversal points occur.
        These are used if we want to copy new classes using the same parameters.
        """
        self.revDist = revDist
        self.revWidth = revWidth
        self.revProminence = revProminence
    
    def _setStatePropreties(self, revDist, revWidth, revProminence,
                                 setCycles, setArea, setSlope):
        """
        Sets the propreties that are used to caculate where reversal points occur.
        These are used if we want to copy new classes using the same parameters.
        """
        self.revDist = revDist
        self.revWidth = revWidth
        self.revProminence = revProminence           
        
        self._setCycles = setCycles
        self._setArea = setArea
        self._setSlope = setSlope   
        
    def _getStatePropreties(self):
        """
        Sets the propreties that are used to caculate where reversal points occur.
        These are used if we want to copy new classes using the same parameters.
        """
        
        return [self.revDist, self.revWidth, self.revProminence,
                self._setCycles, self._setArea, self._setSlope]
        
    def setReversalIndexes(self, revDist = 2, revWidth = None, 
                           revProminence = None, **kwargs):
        """ 
        Finds the location of the reversal points and sets them in the 
        'reversalIndexes'
            
        Parameters
        ----------
        
        revDist : int, optional
            Used to filter reversal points based on the minimal horizontal distance 
            (>= 1) between neighbouring peaks in the x axis. Smaller peaks are 
            removed first until the condition is fulfilled for all remaining peaks.
            The default is 2.
        revWidth : int, optional
            Used to filter reversal points using the approximate width in number of 
            samples of each peak at half it's prominence. Peaks that occur very 
            abruptly have a small width, while those that occur gradually have 
            a big width.
            The default is None, which results in no filtering.  
        revProminence : number, optional
            Used to filter reversal points that aren't sufficently high. Prominence 
            is the desired difference in height between peaks and their 
            neighbouring peaks. 
            The default is None, which results in no filtering.     
        
        """
       
        self.setReversalPropreties(revDist, revWidth, revProminence)
        
        x = self.xy[:,0]
        self.reversalIndexes = data.getCycleIndexes(x, revDist, revWidth, revProminence)
        self.loadProtocol = x[self.reversalIndexes]

    def getReversalxy(self):
        """
        Gets the reversal xy points
        """
        return self.xy[self.reversalIndexes]
    
    def setCycles(self):
        """ Stores all simple cycle objects in the Hysteresis
        """
        xy = self.xy
        indices = self.reversalIndexes
        NIndex = len(indices) - 1
        
        Cycles = [None]*NIndex
        for ii in range(NIndex):
            # I forget why we overshoot the cycles here by one
            Cycles[ii] = SimpleCurve(xy[indices[ii]:(indices[ii+1]+1), :])
                       
        self.cycles:list[SimpleCurve] = Cycles
        self.NCycles = NIndex
        self.NCycles = len(Cycles)
    
    def getCycles(self, Indexes):
        """ Returns a list of cycles given a list of indexes"""
        Cycles = [self.cycles[index] for index in Indexes]
        return Cycles      
    
    def getCycle(self, Index):
        """ Returns a list of cycles given a input index"""
        return self.cycles[Index]

    def setCycleNetAreas(self):
        """ Calculates the net area under each cycle."""
        areas = np.zeros(self.NCycles)
        for ii, vector in enumerate(self.cycles):
            # Skip the last point to avoid counting it twice!
            areas[ii] = vector.getNetArea(endIndex = -1)
            
        self.CycleAreas = areas

    def setNetArea(self):
        """ Sets the net area of the whole Hysteresis """
        self.NetArea = np.sum(self.area)

    def plotCycle(self, Index, showPeaks = False, **kwargs):
        
        """
        Plots the xy values of a single cycle.
        
        Parameters
        ----------
        Index : int
            The cycle to be plotted.
        plotPeaks : bool
            A switch that specifics if the detected peak values should be plotted.
        kwargs : list, optional
            Any additional keyword arguements will be passed to the matplotlib
            plot function. This can be used to custize char appearance.            
        """
        
        Cycle = self.cycles[Index]
        return Cycle.plot(showPeaks = showPeaks, **kwargs)

    def plotCycles(self, cycleIndexes:list = None, showReversals = False, showPeaks = False, 
                    labelReversals:list = None, colorCycles:list = None, **kwargs):
        """
        Plots the xy values of a several cycles on the same figure.
        If no list is provided, every cycle will be plotted.
        
        Parameters
        ----------
        cycleIndexes : list of int
            A list of the cycles to be plotted. By default everything is plotted.
        showReversals : bool
            A switch that specifics if cycle reversal points should be plotted.            
        showPeaks : bool
            A switch that specifics if the detected peak values should be plotted.
        labelReversals : list of int, or 'all'
            A list of the integers which can be used to label certain reversal points
            on the graph. If no lis is provided then no reversal points are labeled.
            If 'all' is provided, then all points will be labeled
        colorCycles : list of str
            A list of matplotlib color labels to use for the cycle color. The cycles
            will follow the colours used, the wrap back to the start of the list if there
            are more cycles than entries provided in the color Cycles.            
        kwargs : list, optional
            Any additional keyword arguements will be passed to the matplotlib
            plot function. This can be used to custize char appearance.            
        """        
        
        xyHys = self.xy
        
        if not colorCycles:
            colorCycles = self.colorCycles
        
        self.showCycles(xyHys, showReversals, showPeaks, self.peakIndexes, 
                        self.reversalIndexes, labelReversals, cycleIndexes)
                    
        return self._plot_curves(self.cycles, cycleIndexes, colorCycles, **kwargs)
    
     
    

    def recalculateCycles(self, revDist = 2, revWidth = None, revProminence = None, **kwargs):
        """
        Calcualtes the cycles again, using the input parameters for distance 
        (number of indexes), width (distance on the x axis), and prominence
        (distance in the y axis).
        
        Peaks are calculated using scipy's find_peaks function.
        
        Parameters
        ----------
            
        revDist : number, optional
            The minimum minimal numbr of indexese (>= 1) in samples between
            neighbouring peaks. Smaller peaks are removed first until the condition
            is fulfilled for all remaining peaks.
        revProminence : number or ndarray or sequence, optional
            Required prominence of peaks. Either a number, ``None``, an array
            matching `x` or a 2-element sequence of the former. The first
            element is always interpreted as the  minimal and the second, if
            supplied, as the maximal required prominence.
        revWidth : number or ndarray or sequence, optional
            Required width of peaks in samples. Either a number, ``None``, an array
            matching `x` or a 2-element sequence of the former. The first
            element is always interpreted as the  minimal and the second, if
            supplied, as the maximal required width.
        
        
        Returns
        -------
        None.

        """
        
        self.setReversalIndexes(revDist, revWidth, revProminence)
        self.setCycles()
        self.setArea()
        
        
    def recalculateCycles_like(self, sampleHysteresis):
        """
        Recalulates the cycles of one hysteresis using another the reversal
        propreties from another hysteresis. Peaks are calculated using scipy's 
        find_peaks function.
        
        Parameters
        ----------
            
        sampleHysteresis : Hysteresis, optional
            The hysteresis to be be used when recalculating the cycle points.
        
        Returns
        -------
        None.

        """        
        revDist = sampleHysteresis.revDist
        revWidth = sampleHysteresis.revWidth
        revProminence = sampleHysteresis.revProminence
       
        self.setReversalIndexes(revDist, revWidth, revProminence)
        self.setCycles()
        self.setArea()        
        
    def recalculateCycles_dist(self, revDist = 2, revWidth = None, 
                               revProminence = None, **kwargs):
        """
        Calcualtes the cycles again, using the input parameters for distance 
        (number of indexes), width (distance on the x axis), and prominence
        (distance in the y axis).
        
        The instead of the xy curve, the secant length between each point on 
        the curve is used to find revesal indexes. 
        Peaks are calculated using scipy's find_peaks function.
        
        Parameters
        ----------
             
        revDist : number, optional
            The minimum minimal numbr of indexese (>= 1) in samples between
            neighbouring peaks. Smaller peaks are removed first until the condition
            is fulfilled for all remaining peaks.
        revProminence : number or ndarray or sequence, optional
            Required prominence of peaks. Either a number, ``None``, an array
            matching `x` or a 2-element sequence of the former. The first
            element is always interpreted as the  minimal and the second, if
            supplied, as the maximal required prominence.
        revWidth : number or ndarray or sequence, optional
            Required width of peaks in samples. Either a number, ``None``, an array
            matching `x` or a 2-element sequence of the former. The first
            element is always interpreted as the  minimal and the second, if
            supplied, as the maximal required width.
        
        Returns
        -------
        None.

        """        
        
        
        try:
            dxy = self.length
        except:
            raise Exception('Length not set yet, set length using .setLength')
                        
        self.reversalIndexes = data.getMaxIndexes(dxy, revDist, revWidth, revProminence)
        self.loadProtocol = self.xy[self.reversalIndexes,0]
        
        self.setCycles()
        self.setArea()           

class SimpleCurve(Curve):
    """ 
    A simple cycle is a curve that doesn't change direction on the X axis, but 
    can change direction in it's Y wzis. The data has a number of peaks, each 
    of which is the largest y value in relation to other points on the curve. 
    This point canbe 
    
    
    Parameters
    ----------
    XYData : array
        The input array of XY date for the curve. A list [x,y] can also be 
        passed into the function.
    peakDist : int, optional
        Used to filter peaks based on the minimal horizontal distance in number 
        of samples (>= 1) between neighbouring peaks in the y axis. 
        Smaller peaks are removed first until the condition is fulfilled for 
        all remaining peaks.
        The default is 2.
    peakWidth : int, optional
        Used to filter peaks points using the approximate width in number of 
        samples of each peak at half it's prominence. Peaks that occur very 
        abruptly have a small width, while those that occur gradually have 
        a big width.
        The default is None, which results in no filtering.  
    peakProminence : number, optional
        Used to filter peaks points that aren't sufficently high. Prominence 
        is the desired difference in height between peaks and their 
        neighbouring peaks. 
        The default is None, which results in no filtering.        
    findPeaks : Boolean, optional
        Used to turn on or off setting the cycle peak points. 
        Can be turned off for performance reasons.
    setArea : Boolean, optional
        Used to turn on or off setting the Area by default. 
        Can be turned off for performance reasons.
    setSlope : Boolean, optional
        Used to turn on or off setting the slopeby default. 
        Can be turned off for performance reasons.   
    
    """
    
    colorCycles = ['C0', 'C1'] # applies to subcycles here
    
    def __init__(self, XYData, findPeaks = False, setSlope = False, setArea = False,
                 peakDist = 2, peakWidth = None, peakProminence = None, **kwargs):
        Curve.__init__(self, XYData, **kwargs)
        
        self._setDirection()
        self._setStatePropreties(findPeaks, setSlope, setArea, 
                                 peakDist, peakWidth, peakProminence)
               
        self.subCycles = None
        if findPeaks == True:
            self.setPeaks()
            self.setSubCycles()
    
        if setSlope == True:
            self.setSlope()
        if setArea == True:
            self.setArea()      

    
    def _setStatePropreties(self, findPeaks, setSlope, setArea, 
                                  peakDist, peakWidth, peakProminence):
        """
        Sets the varabiles that are used to caculate what propreties of the curve
        are run. Used when copying the simple cycle object
        """
        self._findPeaks = findPeaks
        self._setSlope = setSlope
        self._setArea = setArea
        
        self.peakDist = peakDist
        self.peakWidth = peakWidth
        self.peakProminence = peakProminence        

    def _getStatePropreties(self):
        """
        Sets the propreties that are used to caculate where reversal points occur.
        These are used if we want to copy new classes using the same parameters.
        """
        return [self._findPeaks,self._setSlope,self._setArea,
                self.peakDist, self.peakWidth, self.peakProminence]

    def _setDirection(self):
        """
         1 = left to right. -1 = to is right to left

        """
        xdata = self.xy[:,0]
        if xdata[0] <= xdata[-1]:
            self.direction =  1
        else:
            self.direction = -1
           
                
    def setSubCycles(self, peakDist = 2, peakWidth = None, peakProminence = None):
        """
        Finds monotonic "sub-sycles" within each cycle.
        Peaks must be set before subcyles can be set.

        """

        xy = self.xy
        if self.peakIndexes is None:
            self.setPeaks(peakDist, peakWidth, peakProminence)
        indices = self.peakIndexes
        NIndex = len(indices) - 1
        
        SubCycles = [None]*NIndex
        for ii in range(NIndex):
            SubCycles[ii] = MonotonicCurve(xy[indices[ii]:(indices[ii+1]+1), :])
                       
        self.subCycles = SubCycles
        self.NsubCycles = len(SubCycles)
                
    def setSubCyclesArea(self):
        """
        Sets the net area for all SubCycles.
        """
        try:
            SubCycles = self.subCycles
        except:
            raise Exception('No SubCycles not yet set')
            
        for subCycle in SubCycles:
            subCycle._setNetArea        

    def setSubCyclesSlope(self):
        """
        Sets the net slope for all MonotonicCurves.
        """
        try:
            SubCycles = self.subCycles
        except:
            raise Exception('SubCycles not yet set')
            
        for SubCycle in SubCycles:
            SubCycle.setSlope()       
      
    def getSubCycles(self, indexes):
        """
        returns a list of subcycles with the input indexes.

        Parameters
        ----------
        Index : list of int
            The integer index of the sub-cycle you want to return.

        Returns
        -------
        Subcycle
            The output subcycle.

        """        
        
        SubCycles = [self.subCycles[index] for index in indexes]
        return SubCycles      
    
    def getSubCycle(self, index):
        """
        Gets the subcycle with a partcular index.

        Parameters
        ----------
        Index : int
            The integer index of the sub-cycle you want to return.

        Returns
        -------
        Subcycle
            The output subcycle.

        """
        return self.subCycles[index]
    
    def plotSubCycles(self, subCyclesIndexes:list = None, showReversals = False, showPeaks = False,
                      colorCycles:list = None, **kwargs):
        """
        Plots the xy values of a several subcycles on the same figure.
        If no list is provided, every subcycle will be plotted.
        
        Parameters
        ----------
        subCyclesIndexes : list of int
            A list of the subcycles to be plotted.
        showReversals : bool
            A switch that specifics if subcycle reversal points should be plotted.
            It's expected that there are no reveral points in each subcycles'            
        showPeaks : bool
            A switch that specifics if the detected peak values should be plotted.
            
        kwargs : list, optional
            Any additional keyword arguements will be passed to the matplotlib
            plot function. This can be used to custize char appearance.            
        """  
        
        xyMono = self.xy
        
        if not colorCycles:
            colorCycles = self.colorCycles
        
        self.showCycles(xyMono, showReversals, showPeaks, self.peakIndexes, 
                        self.reversalIndexes, None, subCyclesIndexes)
    
        return self._plot_curves(self.subCycles, subCyclesIndexes, colorCycles, **kwargs)
    
    
        
    def recalculatePeaks(self, peakDist = 2, peakWidth = None, peakProminence = None):
        """
        Peaks are calculated using scipy's find_peaks function
        
        Parameters
        ----------
            
        distance : number, optional
            Required minimal horizontal distance (>= 1) in samples between
            neighbouring peaks. Smaller peaks are removed first until the condition
            is fulfilled for all remaining peaks.
        prominence : number or ndarray or sequence, optional
            Required prominence of peaks. Either a number, ``None``, an array
            matching `x` or a 2-element sequence of the former. The first
            element is always interpreted as the  minimal and the second, if
            supplied, as the maximal required prominence.
        width : number or ndarray or sequence, optional
            Required width of peaks in samples. Either a number, ``None``, an array
            matching `x` or a 2-element sequence of the former. The first
            element is always interpreted as the  minimal and the second, if
            supplied, as the maximal required width.
        
        
        Returns
        -------
        None.

        """
        self.setPeaks(peakDist, peakWidth, peakProminence)
        self.setSubCycles()
        self.setArea()  

class MonotonicCurve(Curve):
    """
    A curve that has no changes in it's x axis direction, as well as no changes
    in it's y axis direction.
    """   
    def __init__(self, XYData, **kwargs):
        Curve.__init__(self, XYData, **kwargs)
        
        self._setDirection()
            
    def _setDirection(self):
        """
         1 = neg to pos. -1 = pos to neg
        """
        ydata = self.xy[:,1]
        if ydata[0] <= ydata[-1]:
            self.direction =  1
        else:
            self.direction = -1






import numpy as np
import hysteresis as hys

t = np.linspace(0,4,100)*np.pi
x = np.sin(t)
y = np.cos(t)*t
xy = np.column_stack([x,y])
ty = np.column_stack([t ,y])
tyReversed = np.column_stack([t[::-1] ,y[::-1]+y])

# myHys = hys.Hysteresis(xy)
# resampled = hys.resampledx(myHys, 10)

dx = 0.1
cycle = hys.SimpleCurve(ty, findPeaks=True)
resampledCycle = hys.resampleDx(cycle, dx)

# resampledCycles.plot()
# hys.resample(curve, Nsamples, kwargs)


# cycle1 = hys.SimpleCurve(ty, findPeaks=True)
# cycle2 = hys.SimpleCurve(tyReversed, findPeaks=True)
# curves = [cycle1, cycle2]

# hysWithReversals = hys.concatenate(curves)

def test_resampledx():
    """ Tests if a resampled hysteresis has the correct number of cycles. """
    
    error1 = abs(abs(resampledCycle[10][0]-resampledCycle[11][0])/dx - 1)
    error2 = abs(abs(resampledCycle[-10][0]-resampledCycle[-11][0])/dx - 1)
    assert (error1 < 0.02) and  (error2 < 0.02)

# test_resampledx()





import numpy as np

from scipy.interpolate import interp1d
from hysteresis import data


def defaultSlopeFunction(xy):
    """
    The standard slope function. The slope is undefined for reversal indexes.
    
    For middle points, a centeral finite difference scheme is used, which is 
    O(h**2) accurate for constant steps h.
    
    The slope is equal to the left difference for the end, and
    right difference for the start.
       
    Parameters
    ----------
    xy : Array
        The input xy data in dimension [Ndata, 2] .
        
    Returns
    -------
    The slope of the input cuve at each x point.

    """
    # TODO: Try to replace with numpy.gradient

    npoint = len(xy[:,0])
    if 3 < npoint:
        xyp = xy[2:, :]
        xyn = xy[:-2, :]
        
        dxMid = (xyp[:,0] - xyn[:,0])
        unDefinedIndex = np.array(np.where(dxMid == 0))
        dxMid[unDefinedIndex] = np.max(np.abs(dxMid))
        slopeMid = ((xyp[:,1] - xyn[:,1]) / dxMid)
        slopeMid[unDefinedIndex] = slopeMid[unDefinedIndex - 1]
        
    Startdx = xy[1,0] - xy[0,0]
    Enddx = xy[1,0] - xy[0,0]
    
    slopeStart = 0 
    slopeEnd = 0
    
    if Startdx != 0:
        slopeStart = (xy[1,1] - xy[0,1]) / Startdx
        
    if Startdx != 0:
        slopeEnd = (xy[-1,1] - xy[-2,1]) / Enddx
    
    return np.concatenate([[slopeStart], slopeMid, [slopeEnd]])  
   
def defaultAreaFunction(xy):
    """
    The standard area function. This is an implementation of the midpoint rule.
    
    Parameters
    ----------
    xy : Array
        The input xy data in dimension [Ndata, 2] .

    Returns
    -------
    Area of the xy curve under each point.

    """
    
    # calculate the difference between each point
    dx = np.diff(xy[:,0])    
    yMid = xy[0:-1,1] + np.diff(xy[:,1])/2
    
    areaMid = dx*yMid
    
    # Get the area on either side of the midpoint
    areap = areaMid[1:]
    arean = areaMid[:-1]
    
    # Get the total area of the point
    AreaCenter = ((areap + arean) / 2)
    AreaStart = (areaMid[0]) / 2
    AreaEnd = (areaMid[-1]) / 2   
    
    
    return np.concatenate([[AreaStart], AreaCenter, [AreaEnd]])

def defaultSampleFunction(xy1, xy2):
    """
    The sample function compares how similar two curves are. 
    If they are exactly the same it will return a value of zero.
    The default function returns the average error between each sample point in two arrays of x/y points,
    xy1 and xy2.
    
    Parameters
    ----------
    xy1 : array
        The first input 2D x/y array of points.
    xy2 : array
        The second input 2D x/y array of points.

    Returns
    -------
    float
        The average "distance" between each point on the curve. The output quantity is unitless.
    """

    
    x1 = xy1[:,0]
    x2 = xy2[:,0]
    y1 = xy1[:,1]
    y2 = xy2[:,1]
    
    diff = ((x1 - x2)**2 + (y1 - y2)**2)**(0.5)
    
    return np.sum(diff)/len(x1)


def defaultCombineDiff(diffs):
    """
    This function chooses how to combine the differences between the curves for each cycle.
    The default function returns the average difference for each cycle for a input list of differences
    

    Parameters
    ----------
    diffs : array
        The list of net differences for each input sample curve.

    Returns
    -------
    diffNet : float
        The total difference according too the function.

    """
    
    # The average difference for each curve
    diffNet = np.sum(diffs)/len(diffs)
    
    return diffNet


    
def defaultLengthFunction(xy):
    """
    This function gets the lengths of the line segment between each point in
    a xy curve.
    If they are exactly the same it will return a value of zero.
    The default function returns the average error between each sample point in two arrays of x/y points,
    xy1 and xy2.
    
    Parameters
    ----------
    xy1 : array
        The first input 2D x/y array of points.
    xy2 : array
        The second input 2D x/y array of points.

    Returns
    -------
    float
        The average "distance" between each point on the curve. The output quantity is unitless.
    """

    # xy = self.xy
    # Get the length of each point 
    xybase = xy[:-1]
    xyNext = xy[1:]
        
    dyx = np.sum((xybase- xyNext)**2,1)**(0.5)

    # diff = ((x1 - x2)**2 + (y1 - y2)**2)**(0.5)
    
    return dyx



# -*- coding: utf-8 -*-
"""
Created on Sat May  1 23:49:09 2021

@author: Christian

IT's diffcult to test the GUI functions 


"""


import numpy as np
import hysteresis.plotSpecial.animate as ani

testBase = ani.AnimationBase()
testBase.isPlaying = True
testBase.frames = np.arange(20)

    
def test_Toggle():
    testBase.togglePlay()
    assert testBase.isPlaying == False
    
def test_get_next_frame_incriment():
    frame = testBase._get_next_frame(5)
    assert frame == 6

def test_get_next_frame_loop():
    frame = testBase._get_next_frame(21)
    assert frame == 0







# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt
        
disp = np.loadtxt('UFP_Disp.out',delimiter=' ')
force = np.loadtxt('UFP_RFrc.out',delimiter=' ')
testHys2 = np.column_stack([disp[:,1], -force[:,1]])
DamperHys = hys.Hysteresis(testHys2)


def test_plot_labels(monkeypatch):
    """ Tests if the cycles can plot correctly"""
    
    monkeypatch.setattr(plt, 'show', lambda: None)
    DamperHys = hys.Hysteresis(testHys2)

    DamperHys.plot(showReversals = True, labelReversals = [17,18,19,23])
    plt.close()
    DamperHys.plot(showReversals = True, labelReversals = 'all')
    plt.close()

    assert True == True


def test_plot_Cycles(monkeypatch):
    """ Tests if the cycles can plot correctly"""
    
    monkeypatch.setattr(plt, 'show', lambda: None)
    DamperHys = hys.Hysteresis(testHys2)

    DamperHys.plot(showReversals = True, labelReversals = [17,18,19,23])
    plt.close()
    DamperHys.plotCycles(showReversals = True, labelReversals = [3,6,8])
    plt.close()
    DamperHys.plotCycles([0,1])
    plt.close()

    assert True == True

def test_plot_Area(monkeypatch):
    """ Tests if the area can plot correctly."""
    
    monkeypatch.setattr(plt, 'show', lambda: None)   
    DamperHys = hys.Hysteresis(testHys2)  
    DamperHys.plotArea()

    assert True == True


def test_plot_Slope(monkeypatch):
    """ Tests if the slope can plot correctly"""
    
    monkeypatch.setattr(plt, 'show', lambda: None)   
    DamperHys = hys.Hysteresis(testHys2)  
    slope = DamperHys.slope
    DamperHys.plotSlope()
    plt.close()
    
    assert True == True


    


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt
import scipy



# a noisey triangle signal
np.random.seed(101)
x = np.linspace(0, 1, 1000)*10
triangleBig = scipy.signal.sawtooth(x*2,0.5)
permutate = np.random.normal(0,1,1000)/2
Ynoise = triangleBig + permutate
Ynoise = scipy.signal.savgol_filter(Ynoise,53,2)
trianglexy = np.column_stack((x, Ynoise))


def test_noisey_triangles():
    noiseyTriangles = hys.SimpleCurve(trianglexy, findPeaks = True)
    noiseyTriangles.recalculatePeaks(peakWidth = 100)
    peak2 = noiseyTriangles.peakIndexes
    noiseyTriangles.plot(showPeaks = True)
    
    test1 = len(peak2) == 7
    test2 = peak2[-2] == 794
    assert np.all([test1, test2])
    
    
def test_noisey_triangle_Area():
    noiseyTriangles = hys.SimpleCurve(trianglexy, findPeaks = True)
    noiseyTriangles.setArea()
    area = noiseyTriangles.getNetArea()

    assert (area - -0.22793618247762648) < 10**-5


# -*- coding: utf-8 -*-
"""
Created on Fri May  7 20:15:19 2021

@author: Christian

Notes, currently an experimentla feature. In the future I'd like to
    Add arrow key functionality
    Add bliting to speed up rendering.

"""

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.widgets import Button, Slider
import numpy as np

# Add this function to somewhere else
def init_Animation():
    fig, ax = plt.subplots()
    
    return fig, ax
    
def getAnixy(rawFrames, skipFrames):
    xyAni = rawFrames[::skipFrames,:]
    
    return xyAni

def getAniFrames(x:list, targetdx:float):
    """
    This function is used to get a set of animation frames from a input vector.
    For the input vector x, a single frame will be pulled from every region
    dx. Assumes that x is monotonically increasing.
    
    This can be useful for pre-processing dataif input data has a variable timestep.
    
    No linear inerpolatin is used for intermediate frames.
    
    Parameters
    ----------
    x : list
        the array of input values.
    targetdx : float
        the dx for the output animation..

    Returns
    -------
    np.array
        the output np array of points, with one frame every dx from the 
        target dx.

    """

    NFrames = len(x)   
    NframesOut = []
    
    jj = 0
    for ii in range(NFrames):
        
        while jj*targetdx < x[ii]:
            NframesOut.append(x[ii])
            jj+=1

    return np.array(NframesOut)



class AnimationBase:    
    
    def initAnimation(self):
        self.isPlaying = True
        self.fig, self.ax = init_Animation()
        
    def connectWidgets(self):
        
        """
        Sets up the canvas for wigits. This includes the bottom slider, 
        as well as connecting press and click events.
        """
        plt.subplots_adjust(bottom=0.25)
        
        # Connect clicking
        self.fig.canvas.mpl_connect('button_press_event', self.on_click)
        self.fig.canvas.mpl_connect('key_press_event', self.on_press)
        
        # set up the slider axisd
        self.Sax = plt.axes([0.20, 0.1, 0.7, 0.03])
        self.slide = Slider(self.Sax, "Index", 0, self.Nframes, valinit=0,
                           valstep = 1, valfmt = "%i",  color="green")   
        self.fig.canvas.draw()
        
        # connect the update function to run when the slider changes
        self.slide.on_changed(self.update_line_slider)
        
        # self.canvas.draw_idle()
    def on_press(self, event):
        print(event.key)

    def togglePlay(self):
        """ Turns on or off the animation """
        self.isPlaying = self.isPlaying == False

    # def toggle_pause(self, event, *args, **kwargs):
    def toggle(self, event):
        """
        Toggles playing on or off
        """
        if self.isPlaying == True:
            # self.ani.pause()
            self.ani.event_source.stop()
            self.fig.canvas.draw_idle()
        else:
            # self.ani.resume()
            self.ani.event_source.start()
        self.togglePlay()

    def getClickXY(self, event):
        xclick = event.x
        yclick = event.y
        return xclick, yclick
    
    def stepForward(self, event):
        pass        
    def stepBack(self, event):
        event.key
    def getFrame(self, x):
        pass

    def on_click(self, event):
        """
        Toggles on or off playing if a click event happened within the main
        graph.

        Parameters
        ----------
        event : matplotlib event
            The matplotlib click event class.
        """
        # Check where the click happened
        (xm,ym),(xM,yM) = self.slide.label.clipbox.get_points()
        if xm < event.x < xM and ym < event.y < yM:
            # Event happened within the slider, ignore since it is handled in update_slider
            return
        else:
            self.toggle(event)
    
    
    # Define the update function
    def update_line_slider(self, currentFrame):
        """
        Converts changes in the slider to a frame that can be passed to the 
        update function.
        """
        
        self.aniArtists = self.update(currentFrame)
        self.fig.canvas.draw_idle() 
    
    
    def _get_next_frame(self, currentFrame):
        currentFrame += 1
        if currentFrame >= self.frames[-1]:
            currentFrame = 0
        return currentFrame
    
    def update_slider_widget(self, frame):
       
        """
        Updates the plot based on the current value of the slider.
        Returns a set artists for the animations
        """

        # Find the close timeStep and plot that
        # CurrentFrame = int(np.floor(plotSlider.val))

        # convert the slider value to a time step
        currentTime = self.slide.val
        currentFrame = int(currentTime)
        # CurrentFrame = (np.abs(timeSteps - CurrentTime)).argmin()

        aniframe = self._get_next_frame(currentFrame)
           
        self.slide.set_val(aniframe)
        
    @staticmethod
    def update(self, frame):
        """updates the plot"""
        pass
    
    def animate(self):
        self.initAnimation()
        # set the update function
        if self.widgets == True:        
            self.connectWidgets()
            update = self.update_slider_widget
        else:
            update = self.update
        
        self.ani = FuncAnimation(self.fig, update, self.frames,  
                                 interval=self.interval, blit=False)


class Animation(AnimationBase):
        
    def __init__(self, Curve, pointsPerFrame = 1, skipFrames = 1, 
                 skipStart = 0, skipEnd = 0, interval = 50, widgets = True):
        """
        Creates a animation of the input curve object.
        This is still an experimental feature and may not work as intended 
        for large datasets.

        Parameters
        ----------
        Curve : Hysteresis Curve
            The curve to animate.
        pointsPerFrame : int, optional
            The number of data points to draw per frame. The default is 1.
        skipFrames : int, optional
            THe number of animation frames to skip per input. This reduces
            can be used to reduce the size of large data arrays. The default is
            1, which shows all frames.
        skipStart : int, optional
            Allows the user to skip this many frames at the start.
            Skipped frames are applied after the other frame filters are applied.
            The default is 0, which skips no start frames.
        skipEnd : TYPE, optional
            Allows the user to skip this many frames at the start. 
            Skipped frames are applied after the other frame filters are applied.
            The default is 0, which skips no start frames.
        interval : int, optional
            The target time in ms the frame will be dispalyed for. 
            The default is 50ms.
        widgets : Boolean, optional
            A toggle that allows the user to turn on or off widgets. 
            The default is True, which has the widgets on.

        """

        self.Curve = Curve
        self.pointsPerFrame = pointsPerFrame
        self.skipStart = skipStart
        self.skipEnd = skipEnd
        self.interval = interval        
        self.widgets = widgets
        self.xy = Curve.xy
        
        xyAni   = getAnixy(self.xy, skipFrames)
        self.xyAni = self.skipStartEnd(xyAni, skipStart, skipEnd)
        
        self.Nframes = int(len(self.xyAni) / pointsPerFrame)
        self.frames  = np.arange(self.Nframes)

    def validateData(self):
        pass

    def skipStartEnd(self,xyAni, skipStart, skipEnd):
        
        if skipEnd == 0:
            return xyAni[skipStart:, :]
        else: 
            skipEnd *= -1 
        return xyAni[skipStart:skipEnd, :]

    def update(self, frame):
        """
        Updates the canvas at the given frame.

        """
        points = int(frame*self.pointsPerFrame)
        newXY  = self.xyAni[:points,:]
        line   = self.lines[0]
        line.set_data(newXY[:,0], newXY[:,1])
        # self.fig.canvas.draw_idle() 
        return [line]
    
    def initAnimation(self):
        super().initAnimation()
        line = plt.plot(self.xyAni[:,0], self.xyAni[:,1])[0]
        self.lines = [line]


class JointAnimation(AnimationBase):
    
    def __init__(self, curves, pointsPerFrame = 1, skipFrames = 1, 
                 skipStart = 0, skipEnd = 0, interval = 50, widgets = True):
        """
        Animates several plots on the same graph. All input graphs must have
        the same number of data points.

        Parameters
        ----------
        Curves : list
            The list of hysteresis curves to animate.
        pointsPerFrame : int, optional
            The number of data points to draw per frame. The default is 1.
        skipFrames : TYPE, optional
            THe number of animation frames to skip per input. This reduces
            can be used to reduce the size of large data arrays. The default is
            1, which shows all frames.
        skipStart : int, optional
            Allows the user to skip this many frames at the start.
            Skipped frames are applied after the other frame filters are applied.
            The default is 0, which skips no start frames.
        skipEnd : TYPE, optional
            Allows the user to skip this many frames at the start. 
            Skipped frames are applied after the other frame filters are applied.
            The default is 0, which skips no start frames.
        interval : int, optional
            The target time in ms the frame will be dispalyed for. 
            The default is 50ms.
        widgets : Boolean, optional
            A toggle that allows the user to turn on or off widgets. 
            The default is True, which has the widgets on.

        Returns
        -------
        None.

        """


        super().__init__()
        
        
        self._validateCurves(curves)
        
        self.pointsPerFrame = pointsPerFrame   
        self.skipFrames = skipFrames        
        self.skipStart  = skipStart        
        self.skipEnd    = skipEnd        
        self.interval   = interval
        self.widgets    = widgets

        self.Curves  = curves
        self.Ncurves = len(curves)
        
        xyAni = getAnixy(curves[0].xy, skipFrames)
        self.Nframes = int(len(xyAni) / pointsPerFrame)
        self.frames  = np.arange(self.Nframes)        
    
    def _validateCurves(self, curves):
        Lcurve = len(curves[0])
        ii = 1
        for curve in curves[1:]:
            if len(curve) != Lcurve:
                raise Exception('Curves must all have the same number of datapoints.') 
            ii +=1
    
    
    def initAnimation(self):
        super().initAnimation()

        self.lines = []
        self.xyCurves = [None]*self.Ncurves
        for ii in range(self.Ncurves):
            xy    = self.Curves[ii].xy
            xyAni = getAnixy(xy, self.skipFrames)
            
            self.xyCurves[ii] = xyAni
            line = plt.plot(xyAni[:,0], xyAni[:,1])[0]
            self.lines.append(line)    # def initAnimation(self):        
            
    def update(self, frame):
        points = int(frame*self.pointsPerFrame)
        lines = [None]*self.Ncurves
        for ii in range(self.Ncurves):

            tempXY = self.xyCurves[ii]
            # print()
            
            newXY = tempXY[:points,:]
            line = self.lines[ii]
            line.set_data(newXY[:,0], newXY[:,1])
            lines[ii] = line
        
        self.aniArtists = lines
        return self.aniArtists
    
    # def animate(self):
    #     self.initAnimation()
    #     if self.widgets == True:        
    #         self.connectWidgets()
    #         update = self.update_slider_widget
    #     else:
    #         update = self.update  
    #     self.ani = FuncAnimation(self.fig, update, self.frames,
    #                              interval=50, blit=False)   
    
    
    
    
    
    
 


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt

        
disp = np.loadtxt('UFP_Disp.out',delimiter=' ')
force = np.loadtxt('UFP_RFrc.out',delimiter=' ')
testHys2 = np.column_stack([disp[:,1], -force[:,1]])


def test_plot_Slope():
    """ Tests if the slope is the correct value"""

    DamperHys = hys.Hysteresis(testHys2)  
    slope = DamperHys.slope
    
    assert abs(slope[-2] - 40499.99999999157) < 10**-5
    

    
def test_net_Area():
    """ Tests if the net area is correct."""
    
    DamperHys = hys.Hysteresis(testHys2)
    NetArea = DamperHys.getNetArea()

    assert abs(NetArea - 12390.79659964981) < 10**-5
    
    
def test_reversalIndexes():
    """ Tests if the reversalIndexe is correct."""
    
    DamperHys = hys.Hysteresis(testHys2)
    reversalIndexes = DamperHys.reversalIndexes
    
    assert reversalIndexes[15] == 492
    


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt
from  scipy.signal import sawtooth
from scipy.interpolate import interp1d

x = np.linspace(0, 1, 1000)*10
triangleSmall = sawtooth(x*20,0.5)/7
trianglexy = np.column_stack((x,triangleSmall))
smallTriangles = hys.SimpleCurve(trianglexy, findPeaks = True)





def test_plot_peaks(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)
    smallTriangles.setPeaks()
    smallTriangles.plot(showPeaks = True)
    plt.close()
    assert True == True


def test_plot_SubCycles(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)
    smallTriangles.plotSubCycles()
    plt.close()

    assert True == True


def test_plot_Slope(monkeypatch): 
    monkeypatch.setattr(plt, 'show', lambda: None)

    smallTriangles.setSlope()
    smallTriangles.plotSlope()
    plt.close()
    assert True == True

def test_Slope():
    smallTriangles.setSlope()
    slope = smallTriangles.slope
    assert abs(slope[-1] - -1.8189136353358175) < 10**-5

def test_Area():
    smallTriangles.setArea()
    Anet = smallTriangles.getNetArea()
    assert abs(Anet - 0.004982994450842999) < 10**-8





# smallTriangles.setArea()
# Anet = smallTriangles.getNetArea()
# 0.004982994450842999


from .curve import Hysteresis, SimpleCurve, MonotonicCurve
from .baseFuncs import concatenateHys, concatenate, removeNegative
from .resample import resample, resampleDx, resampleRegion

from .compare import compareCycle, compareHys
from .protocol import exandHysTrace, createProtocol
from .climate import SeasonalCurve
from .envelope import getBackboneCurve, getAvgBackbone, fitEEEP




# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt
import scipy
from scipy.interpolate import interp1d

x = np.linspace(0, 1, 1000)*10
triangleSmall = scipy.signal.sawtooth(x*20,0.5)/10
triangleBig = scipy.signal.sawtooth(x,0.5)
triangle = triangleSmall + triangleBig
Y = x**2


def test_plot_RemoveNeg(monkeypatch):
    """ Tests if the remove Negative attribute runs"""
    
    monkeypatch.setattr(plt, 'show', lambda: None)
    

    trianglexy = np.column_stack((triangle,Y))
    smallTriangles = hys.Hysteresis(trianglexy)
    smallTriangles.recalculateCycles(peakProminence=.2)
    
    cycle1 = smallTriangles.getCycle(0)
    
    xy = cycle1.xy
    x = xy[:,0]
    y = xy[:,1]
    
    direction = cycle1.direction
    difference = np.append(0, np.diff(x))
    
    condtion = np.where(0 <= difference*direction)
    
    plt.subplots()
    plt.plot(x[condtion], y[condtion])
    
    New = hys.removeNegative(cycle1)
    New.plot()

    assert True == True


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt
        
Wall_exp = np.loadtxt('Ts1_Experiment_Shear.csv',delimiter=',')
Wall_analysis_disp = np.loadtxt('Ts1_Load_Dsp.out',delimiter=' ')[1:, 1]
Wall_analysis_shear =np.loadtxt('Ts1_Wall_Reaction.out',delimiter=' ')[1:, 1]

Wall_exp_xy = np.column_stack([Wall_exp[:,0], Wall_exp[:,1]/1000])
Wall_anal_xy = np.column_stack([Wall_analysis_disp, -Wall_analysis_shear/1000])

def getHys():
    ExpHys = hys.Hysteresis(Wall_exp_xy[8300:,:])
    AnalHys = hys.Hysteresis(Wall_anal_xy[500:,:])
    
    return ExpHys, AnalHys

def test_Plot(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)
    
    ExpHys,AnalHys= getHys()
    
    ExpHys.plot(True)
    plt.close()
    AnalHys.plot(True) 
    plt.close()
    
    assert True == True

def test_Plot_Index(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)
    ExpHys,AnalHys= getHys()
    
    ExpHys.plotVsIndex(True)
    plt.close()
    AnalHys.plotVsIndex(True)
    plt.close()
    AnalHys.plotLoadProtocol(comparisonProtocol = ExpHys.loadProtocol)
    plt.close()

    assert True == True

def test_Plot_LoadProtocol(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)
    
    ExpHys,AnalHys= getHys()
    ExpHys.plotLoadProtocol()
    plt.close()
    AnalHys.plotLoadProtocol(comparisonProtocol = ExpHys.loadProtocol)
    plt.close()
    assert True == True

def test_ReCalc():

    
    ExpHys, AnalHys= getHys()
    AnalHys.recalculateCycles(revProminence = 0.005)

    assert True == True
    
def test_compare_loadProt(monkeypatch):
    
    ExpHys,AnalHys= getHys()
    AnalHys.recalculateCycles(revProminence = 0.0045)
    ExpHys.recalculateCycles(revProminence = 0.035*4.1/100/2)
    test1 = len(AnalHys.loadProtocol) == len(ExpHys.loadProtocol)

    assert test1 == True

# def test_compare():
def test_compare(monkeypatch):
    
    ExpHys,AnalHys= getHys()
    AnalHys.recalculateCycles(revProminence = 0.0045)
    ExpHys.recalculateCycles(revProminence = 0.035*4.1/100/2)
    
    AnalHys = hys.resample(AnalHys, 10)
    ExpHys = hys.resample(ExpHys, 10)
    
    Diff, Diffs = hys.compareHys(AnalHys, ExpHys)
    
    test1 = abs(Diff / 6.79762) - 1 < 0.001
    
    assert test1 == True

test_compare(None)







# # Now we can resample and compare the curves!
# AnalHysDx = hys.reSample(AnalHys, 10)
# ExpHysDx = hys.reSample(ExpHys, 10)
# AnalHysDx.plotLoadProtocol(comparisonProtocol = ExpHysDx.loadProtocol)
# AnalHysDx.plot(True)
# ExpHysDx.plot(True)

# Diff, Diffs = hys.CompareHys(AnalHysDx, ExpHysDx)
    
    
#     assert True == True

ExpHys,AnalHys= getHys()

ExpHys.plot(True)
# plt.close()
AnalHys.plot(True) 
plt.close()



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian
"""
import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt
        
disp = np.loadtxt('UFP_Disp.out',delimiter=' ')
force = np.loadtxt('UFP_RFrc.out',delimiter=' ')
testHys2 = np.column_stack([disp[:,1], -force[:,1]])


# def test_plot_Slope():
""" Tests if the slope is the correct value"""
def test_cum_Disp():
    DamperHys = hys.Hysteresis(testHys2)  
    
    LP = DamperHys.loadProtocol
    # The total area contained by the hystresis, including end cycles
    cumDisp = DamperHys.getNetCumDisp()
    LPcumDisp = np.sum(np.abs(LP[:-2]))*2 + LP[-2] + (LP[-2] - LP[-1])
    
    assert abs((cumDisp - LPcumDisp)) < 10**-8

    
# test_cum_Disp()


import numpy as np
import hysteresis as hys


t = np.linspace(0,4,100)*np.pi
x = np.sin(t)
y = np.cos(t)*t
xy = np.column_stack([x,y])
ty = np.column_stack([t ,y])
tyReversed = np.column_stack([t[::-1] ,y[::-1]+y])

myHys = hys.Hysteresis(xy)
resampled = hys.resample(myHys, 10)

cycles = hys.SimpleCurve(ty, findPeaks=True)
resampledCycles = hys.resample(cycles, 5)

cycle1 = hys.SimpleCurve(ty, findPeaks=True)
cycle2 = hys.SimpleCurve(tyReversed, findPeaks=True)
curves = [cycle1, cycle2]

hysWithReversals = hys.concatenate(curves)

def test_Hysteresis_lengths():
    """ Tests if a resampled hysteresis has the correct number of cycles. """
    assert len(resampled.cycles[0]) == 10 and len(resampled.cycles[2]) == 10

def test_SimpleCurve_lengths():
    """ Tests if a resampled SimpleCurve has the correct number of cycles. """
    assert len(resampledCycles.subCycles[0]) == 5 and len(resampledCycles.subCycles[2]) == 5


def test_compare_basic():
    
    # myHys.plot(plotCycles = True)
    smallHys = hys.resample(myHys, 10)
    
    out = hys.compareHys(smallHys,myHys)
    
    assert out[0] == 0
    

def test_combined_length():
    output = hys.resample(hysWithReversals, 10)
    Nout = [len(curve) for curve in output.cycles]
    check2 = sum(Nout) == 20
    assert (len(output) == 19) and check2
    




import numpy as np
import matplotlib.pyplot as plt


def initializeFig(xlim = [], ylim= []):
    """
    A default function that initializes a figure based on a set of incoming
    x and y limits

    Parameters
    ----------
    xlim : list
        The x limits in [xmin, xmax].
    ylim : list
        The y limits in [xmin, xmax].

    """

    fig, ax = plt.subplots()

    if len(xlim) != 0 :
        ax.set_xlim(xlim[0], xlim[1])

    if len(ylim) != 0 :
        ax.set_ylim(ylim[0], ylim[1])
               
    return fig, ax

def defaultShowCycles(xy, showReversals, showPeaks, peakIndexes = None, 
                      reversalIndexes=None,  labelCycles = None, cycleIndicies = None):
    """
    A function that plots the location of the reversal points and peaks for a
    curve object.

    Parameters
    ----------
    x : array
        The input x values as an numpy array.
    y : array
        The input y values as an numpy array.
    plotCycles : bool
        A switch that specifics if cycle reversal points should be plotted.
    plotPeaks : bool
        A switch that specifics if peak values should be plotted.
    labelCycles : list or 'all', optional
        A list of the cycles to be labled. 
        A value of 'all' can also be specified, in which case all cucles will be plotted.
        The default is [], which labels no cycles
    Cycles : kist, optional
        A list of the cycles to be plotted. If not specified, all values will 
        be plotted. The default is NOne, which plots no labels cycles.

    """
    
    x = xy[:,0]
    y = xy[:,1]
    
    if labelCycles is None:
        labelCycles = []
    
    if cycleIndicies is None:
        cycleIndicies = []    
        
    # Plot cycles as x only
    if showPeaks == True:
        
        inds = peakIndexes
        if peakIndexes is None:
            raise Exception('No peaks have been set')
            
        PeakX = x[inds]
        PeakY = y[inds]

        line2  = plt.plot(PeakX, PeakY, "+")
        # plt.title('Peak Index y values')    # Plot cycles as x only
        
    if showReversals == True:
        
        if reversalIndexes is None:
            raise Exception('No peaks have been set')
        # TODO: update marker based on direction?
        
        # if only certain cycles have been asked for we remove all indexes other indexes
        if cycleIndicies != []:
            markerIndexes = reversalIndexes[cycleIndicies]
            # only label Cycles that are in the cycles asked for
            labelCycles = np.array([label for label in labelCycles if label in cycleIndicies])
        else:
            # Otherwise we plot all
            markerIndexes = reversalIndexes
        
        # Indexes = Indexes[]
        markerX = x[markerIndexes]
        markerY = y[markerIndexes]         

        # Plot the desired indexes
        line2  = plt.plot(markerX, markerY, "x")        

        # If the cycles need to be labeled,
        if str(labelCycles) == 'all':
            # skip the first and last cycles
            labelIndexes = np.arange(0,len(markerIndexes))
            labelX = x[reversalIndexes]
            labelY = y[reversalIndexes]
        else:
            labelIndexes = labelCycles
            labelX = x[reversalIndexes[labelCycles]]
            labelY = y[reversalIndexes[labelCycles]] 

            
        for ii in range(len(labelIndexes)):
            Annotate = plt.annotate(labelIndexes[ii], [labelX[ii],labelY[ii]], xytext=(-1, 5), textcoords = 'offset points')


def defaultPlotFunction(xy, showReversals, showPeaks, peakIndexes = None, 
                      reversalIndexes=None, labelCycles = None, **kwargs):
    """
    Parameters
    ----------
    xy : 2D array [Npoints, 2]
        The input x/y values as an numpy array, in order x,y and shape [Npoints, 2].
        e.g. x = xy[:,0]
    y : array
        The input y values as an numpy array.
    plotCycles : bool
        A switch that specifics if cycle reversal points should be plotted.
    plotPeaks : bool
        A switch that specifics if peak values should be plotted.
    labelCycles : list or 'all', optional
        A list of the cycles to be labled. 
        A value of 'all' can also be specified, in which case all cucles will be plotted.
        The default is [], which labels no cycles
    Cycles : kist, optional
        A list of the cycles to be plotted. If not specified, all values will 
        be plotted. The default is [], which plots all cycles.
    """
    
    x = xy[:,0]
    y = xy[:,1]
          
    line, = plt.plot(x, y, **kwargs)
       
    defaultShowCycles(xy, showReversals, showPeaks, peakIndexes, reversalIndexes, labelCycles)
    
    return line
    



    




# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 23:47:36 2020

@author: Christian

This tests if a simple cicle works
"""
import hysteresis as hys
import numpy as np
import matplotlib.pyplot as plt


def MakeCircle():
    basePoints = np.linspace(0,1,1000)*2*np.pi
    testCirclex = np.sin(basePoints)
    testCircley = np.cos(basePoints)
    Circlexy = np.column_stack((testCirclex, testCircley))

    Circle = hys.Hysteresis(Circlexy)

    return Circle

def test_Circle_Area():
    """ Test the net area of the circle """
    Circle = MakeCircle()
    NetArea = Circle.getNetArea()
    
    assert abs(NetArea - np.pi) < 0.0001
    
def test_Circle_SLope():
    """ Test the slope area of the circle """
    Circle = MakeCircle()
    Cycle1 = Circle.getCycle(0)
    Cycle1.setSlope()
    x = Cycle1.xy[:,0]
    ind = np.argmin( abs(x - 0.5))
    testSlope = -x[ind] / (1 - x[ind]**2)**0.5
    
    Slope2 = Cycle1.slope[ind]
    
    assert abs(testSlope - Slope2) < 10**-5

def test_Circle_Plot(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)
    
    Circle = MakeCircle()
    Circle.plotCycles()
    Circle.plot(showReversals=True)
    Circle.plotSlope(showReversals=True)
    
    assert True == True

def test_Circle_Subvector_Plot(monkeypatch):
    monkeypatch.setattr(plt, 'show', lambda: None)
    
    Circle = MakeCircle()
    Vector1 = Circle.cycles[0]
    Vector2 = hys.resample(Vector1, 30)
    Vector3 = hys.resample(Circle, 10)
    
    Vector1.plot()
    Vector2.plot()
    Vector3.plot(True)
    
    assert True == True




# -*- coding: utf-8 -*-
"""
Created on Sat Aug 31 15:13:13 2019
@author: Christian
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
from scipy.interpolate import interp1d


def getIntersections(vectorY, tol = 1e-6):
    """
    Returns the indicies where the the input vector is close to zero.

    Parameters
    ----------

    vectorY : list | nd.array
        The inpout vector of points. 
    tol
        The tolerance of how close a value is to zero to be considered. 
        
    Returns
    -------
    inds : list[int]
        The list of points intersection points close to zero.

    """
        
    # Get values that are within a tolerance of zero but don't pass it.
    indsPosTol = vectorY < tol
    indsNegTol = -tol < vectorY 
    
    # Get values within the positive and negative range
    indsTol = indsPosTol*indsNegTol
    
    # Set these noisey values equal to zero
    vectorY[indsTol] = 0
    
    signs = np.sign(vectorY)

    # Manually patch "bad" indexes
    
    zeroIndsToCheck = np.argwhere(indsTol == True).flatten()

    if len(zeroIndsToCheck) != 0:
        
        # zeroIndsToCheck = zeroIndsToCheck[0]
        # remove the start index
        endInd = len(indsTol) - 1
        if 0 == zeroIndsToCheck[0]:
            zeroIndsToCheck = zeroIndsToCheck[1:]
            
        # remove the end index, if there are any more indexes to check.
        if (len(zeroIndsToCheck) != 0) and endInd == zeroIndsToCheck[-1]:
            zeroIndsToCheck = zeroIndsToCheck[:-1]
        
        for ind in zeroIndsToCheck:
            
            sleft  = signs[ind - 1]
            sright = signs[ind + 1]
            
            if sleft != sright:                
                # If the final two items are x,0,0, set them to x,x,0
                # This generally looks better
                if sright == 0:
                    signs[ind] = sleft
                else:
                    signs[ind] = sright

    diff  = np.diff(signs)
    indsInter = np.argwhere(diff).flatten() 

    return indsInter


def getCycleSubVector(vectorX, vectorY, Index1, Index2, Nsample):
    """
      
    
    This function takes a input x y curve, then returns a linearlized curve
    between two indicies
    
    

    Parameters
    ----------
    VectorX : Array
        The input X vector.
    VectorY : Array
        The input Y vector.
    Index1 : int
        The first index we want to calculate values between.
    Index2 : int
        The second index we want to calculate values between.
    Nsample : int
        The desired number of data points between the two vectors.

    Returns
    -------
    xSample : TYPE
        The x points of the sample curve.
    ySample : TYPE
        The y points of the sample curve.

    """
    #TODO consider renaming to interpolate subvector! Right now this samples
    # the vector by default    
    
    x1 = vectorX[Index1]
    x2 = vectorX[Index2]
    
    
    TempDataX = vectorX[Index1:(Index2+1)]
    TempDataY = vectorY[Index1:(Index2+1)]
    xSample = np.linspace(x1,x2,Nsample)
    
    InterpFunction = interp1d(TempDataX, TempDataY)
    ySample = InterpFunction(xSample)
          
    return xSample,ySample


def getMaxIndexes(VectorX, peakDist = 2, peakWidth = None, 
                     peakProminence = None, **kwargs):
    MaxIndexes,_ = find_peaks(VectorX, height = (None, None), distance = peakDist, 
                            width = peakWidth, prominence = peakProminence, **kwargs)

    Nindex = len(MaxIndexes) + 2
    Indexes = np.zeros(Nindex, dtype = int)
    Indexes[0] = 0
    Indexes[-1] = len(VectorX) - 1
    
    Indexes[1:-1] = MaxIndexes

    return Indexes




def _findOrder(L1, L2, MinIndexes, MaxIndexes):
    # We check the order, starting with the minimum number of possibilities
    # if there are no minimus, order doesn't matter
    # If the first index is a minimum, order = 1, otherwise order = 2
    if L1 == 0 and L2 == 0: # Edge case 1: Curve is monotonic - order doesn't matter
        order = 1
    elif L1 == 0: # Edge case 2: There are no intermediate min points and one intermediate max
        order = 2
    elif L2 == 0: # Edge case 3: There are no intermediate maxi points and one intermediate min
        order = 1
    elif MinIndexes[0] < MaxIndexes[0]: 
        order = 1
    else:
        order = 2
    return order



def getCycleIndexes(VectorX, peakDist = 2, peakWidth = None, 
                     peakProminence = None, **kwargs):
    """
    This function finds the index where there is areversal in the XY data. 
    You may need to adjust the find peaks factor to get satisfactory results.
    Built on the scipy find peaks funciton
    

    Parameters
    ----------
    VectorX : 1D array
        Input X Vector.
    VectorY : 1D array
        Input Y Vector. This is only used if we want to plot the values.
    CreatePlot : Boolean
        This switch specifies whether or not to display the output.plot
    peakDist : int, optional
        Required minimal horizontal distance (>= 1) in samples between
        neighbouring peaks. Smaller peaks are removed first until the condition
        is fulfilled for all remaining peaks.
        The default is 2.
    peakWidth : int, optional
        The approximate width in number of samples of the peak at half it's prominence.
        Peaks that occur very abruptly have a small width, while those that occur
        gradually have a big width.
    peakProminence : number, optional
        Used to filter out peaks that aren't sufficently high. Prominence is 
        the desired difference in height between peaks and their neighbouring peaks. 
        The default is None, which results in no filtering.

    Returns
    -------
    Indexes : Arrays
        Returns the arrays at which reversals occur.

    """
       
    # We find the intermediate peak values
    # We use height = 0 to select only positive or negative peaks.
    MaxIndexes,_ = find_peaks(VectorX, height = (None, None), distance = peakDist, 
    # MaxIndexes,_ = find_peaks(VectorX, height = (0, None), distance = peakDist, 
                            width = peakWidth, prominence = peakProminence)
    
    MinIndexes,_ = find_peaks(-VectorX, height = (None, None), distance = peakDist, 
                            width = peakWidth, prominence = peakProminence)    
        
    # We store define an array that will later be used to store the
    # the max and min values in an array of indexes
    Nindex = len(MaxIndexes) + len(MinIndexes) + 2
    Indexes = np.zeros(Nindex, dtype = int)
       
    # Define first and last point
    Indexes[0] = 0
    Indexes[-1] = len(VectorX) - 1
    
    # if only one point exists   
    L1 = len(MinIndexes)
    L2 = len(MaxIndexes)
    
    if abs(L1 - L2) > 1:
        raise Exception('There are', L1, 'minimums which is more than one than ', L2, ' maximums. There are likely repeated peaks.')
    
    # 1 means min first, 2 means max first
    order = _findOrder(L1, L2, MinIndexes, MaxIndexes)
    
    if order == 1:
        Indexes[1:-1:2] = MinIndexes
        if L2 !=0:
            Indexes[2:-1:2] = MaxIndexes
    else:
        Indexes[1:-1:2] = MaxIndexes
        if L1 !=0:
            Indexes[2:-1:2] = MinIndexes      
 
    
    return Indexes


def linearInterpolation(x1, x2, y1, y2, x):
    """
    A default linear interpolation functions for the value x between two points
    p1 = (x1,y1) and p2 = (x2,y2). 

    Parameters
    ----------
    x : float
        The input x value

    Returns
    -------
    y : float
        the output y value.

    """

    dx = (x2 - x1)
    if dx == 0:
        y = y1
    else:
        y = ((y2 - y1) / (dx))*(x - x1) + y1
    
    return y
   





def sampleData(ExperimentX, ExperimentY, AnalysisX, AnalysisY, Nsample = 10, 
               peakDist = 2, peakwidth = None, Norm = None):
    """
        
    This functions samples two data sets that undergo a cyclic, or monotonic 
    response. The function automatically detects reversals in the data. 
    Adjustments may be necessary to find peaks in the data 
    
    The Experiment and Analysis must have the same number of cycles

    Parameters
    ----------
    ExperimentX : Array
        The X values from the experiment dataset.
    ExperimentY : Array
        The Y values from the experiment dataset.
    AnalysisX : Array
        The X values from the analysis dataset.
    AnalysisY : Array
        The Y values from the analysis dataset.

    Raises
    ------
    Exception
        If the datasets don't have the same number of cycles, the sampling
        doesn't work

    Returns
    -------
    Rnet : float
        Returns a the sampled value 

    """
       
    # We get the indicies where the reversal happens.
    
    ExperimentIndicies = getCycleIndexes(ExperimentX, Nsample, peakDist, peakwidth, ExperimentY)
    AnalysisIndicies = getCycleIndexes(AnalysisX, Nsample, peakDist, peakwidth, AnalysisY)
    
    # We check that both curves have the same number of indicies
    NIndex = len(ExperimentIndicies) - 1
    NIndex_2 = len(AnalysisIndicies) - 1
    R= np.zeros(NIndex)
    
    # If they don't we create a error
    if NIndex!=NIndex_2:
        raise Exception('The experiment and Analysis have a different number of cycles')
        
    # For each index, we loop through 
    for ii in range(NIndex):
        # We get define the arrays for the x and y coordinants of the sub-vector
        Ex = np.zeros(NIndex)
        Ey = np.zeros(NIndex)
        Ax = np.zeros(NIndex)
        Ay = np.zeros(NIndex)
        
        # We get the subvector values between the indicies
        [Ex,Ey] = getCycleSubVector(ExperimentX , ExperimentY, ExperimentIndicies[ii], ExperimentIndicies[ii+1], Nsample)
        [Ax,Ay] = getCycleSubVector(AnalysisX , AnalysisY, AnalysisIndicies[ii], AnalysisIndicies[ii+1], Nsample)
        
        # We sample each point on the curve using the difference between the
        # two points
        if Norm == None:
            R[ii] = np.sum(((Ey-Ay)**2 + (Ex-Ax)**2)**0.5)
        else:
            R[ii] = Norm(Ex, Ey, Ay, Ax)
        
    Rnet = np.sum(R)
    
    return Rnet


def sampleMonotonicData(ExperimentX, ExperimentY, AnalysisX, AnalysisY,
                        Nsample = 10, Norm = None):
    """
    This functions works the same way as the cyclic data function
    
    This functions samples two data sets of data and returns a 
    
    The Experiment and Analysis must have the same number of cycles

    Parameters
    ----------
    ExperimentX : Array
        The X values from the experiment dataset.
    ExperimentY : Array
        The Y values from the experiment dataset.
    AnalysisX : Array
        The X values from the analysis dataset.
    AnalysisY : Array
        The Y values from the analysis dataset.

    Returns
    -------
    Rnet : float
        Returns a the sampled value 

    """
    
    Nindex = len(ExperimentX)
    R= np.zeros(Nsample)
    
    for ii in range(Nsample):
        Ex = np.zeros(Nsample)
        Ey = np.zeros(Nsample)
        Ax = np.zeros(Nsample)
        Ay = np.zeros(Nsample)
        
        
        [Ex,Ey] = getCycleSubVector(ExperimentX , ExperimentY, 0, Nindex, Nsample)
        [Ax,Ay] = getCycleSubVector(AnalysisX , AnalysisY, 0, Nindex, Nsample)
        
        R[ii] = np.sum(((Ey-Ay)**2 + (Ex-Ax)**2)**0.5)
        if Norm == None:
            R[ii] = np.sum(((Ey-Ay)**2 + (Ex-Ax)**2)**0.5)
        else:
            R[ii] = Norm(Ex, Ey, Ay, Ax)        
    Rnet = np.sum(R)
    
    return Rnet





# =============================================================================
# Unused and to be removed
# =============================================================================


def shiftDataFrame1(Samplex, Sampley, Targetx):
    """
    
    SciPy's ID interpolate baseically does what this does, it probably makes
    more sense to use that instead.
    
    This functions shifts x/y of a sample vector to a target x vector.
    Intermediate values are found using linearlizaton
    Both the input and putput data must be a function.
    
    The sample point MUST be within the bounds
    
              sample x/y data
              x1___x2____u1___x2_________x5_________x6
              y1___x2____x1___x2_________x5_________x6
    
              Targert x data
              x1_____x2___x3____x4___x5______x6__x7

    Returns:
              y1_____y2___y3____y4___y4______y5__y7
              
    # We will need to break if 
    Parameters
    ----------
    Samplex : Array 1d.
        The x values for the sample curves we want to shift into.
    Sampley : Array 1d.
        The y values for the sample curves we want to shift into.
    Targetx : Array 1d.
        The X vector we want to shift the target into.

    Returns
    -------
    Targety : Array 1d.
        The Y vector we shifted  the target into.

    """
    
        
    # 
    SampleRate = 1    
    
    # Get number of samples needed
    Nsamples = len(Targetx)
    Targety = np.zeros(Nsamples)
    
    MinData = np.min(Samplex)
    MaxData = np.max(Samplex)
    
    # Define indexes
    ii = 0
    jj = 1
    x1 = Samplex[0]
    y1 = Sampley[0]
    x2 = Samplex[jj]
    y2 = Sampley[jj]
    Currentmax = max(x1,x2)
    Currentmin = min(x1,x2)
    
    while (ii < Nsamples):
        
        #Get the target Point
        x = Targetx[ii]
        
        if x < MinData:
            print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            print("Target is less than minimum bounds")
            print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            #Targety = None
            break
        if x > MaxData:
            print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            print("Target is greater than maximum bounds")
            print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            #Targety = None
            break
        
        # We look to linearly interpolate between the two points
        
        # We check if x is within the range of the current sample points
        # if it isn't, we check the next data points
        while ((x < Currentmin) or (Currentmax < x)):
            jj += 1
            #print(jj)
            x1 = x2
            y1 = y2
            x2 = Samplex[jj]
            y2 = Sampley[jj]

            Currentmax = max(x1,x2)
            Currentmin = min(x1,x2)
               
        
        # Interpolate
        Targety[ii] = linearInterpolation(x1,x2,y1,y2,x)
        
        ii+=1
        
    return Targety


"""
TODO
    Consider removing ShiftDataFrame - I don't think this us used anymore
"""




import hysteresis.curve as hys
import numpy as np


def makeCurve(f):
    "lamda function that creates all curves"
    xdata = np.linspace(0,4,1001)
    y = f(xdata)
    xy = np.column_stack([xdata, y])
    Curve = hys.Curve(xy)   
    return Curve    

def function1(x):
    return x + 2

def function2(x):
    return x**3 - 3*x**2 + 3

def function3(x):
    return 3 + (np.e)**x  
    
    
def test_Curve1_Area():
    Curve = makeCurve(function1)
    Curve.setArea()
    A = Curve.getNetArea()

    assert abs(A - 16) < 0.0001

def test_Curve2_Area():
    Curve = makeCurve(function2)
    Curve.setArea()
    A = Curve.getNetArea()
    
    assert abs(A - 12) < 0.0001
    
def test_Curv3_Area():
    Curve = makeCurve(function3)
    Curve.setArea()
    A = Curve.getNetArea()

    assert abs(A - (np.e**4 + 11)) < 0.0001



# -*- coding: utf-8 -*-
"""
Created on Sat May  1 23:49:09 2021

@author: Christian
"""


import numpy as np
import hysteresis.plotSpecial.animate as ani

import matplotlib.pyplot as plt
import matplotlib.lines as lines
import hysteresis as hys
import scipy

np.random.seed(101)
x = np.linspace(0, 1, 101)*10
y = np.sin(x)
trianglexy = np.column_stack((x, y))

test = hys.Hysteresis(trianglexy)
# testBase = ani.AnimationBase()
# testBase.initAnimation()
skipStart  =3
skipEnd = 3
myAnimation = ani.Animation(test, skipStart = 3, skipEnd = 3)
# Animation
# myAnimatino.skipStartEnd


def test_skip():
    
    xy = myAnimation.skipStartEnd(trianglexy, skipStart, skipEnd)
    assert len(xy) == 101 - skipStart - skipEnd

def test_skip_start():
    
    xy = myAnimation.skipStartEnd(trianglexy, skipStart, 0)
    assert len(xy) == 101 - skipStart
    
def test_update(monkeypatch):
    """ Tests if the cycles can plot correctly"""
    
    monkeypatch.setattr(plt, 'show', lambda: None)
    myAnimation.initAnimation()
    line, = myAnimation.update(10)
    test1 = isinstance(line, lines.Line2D)
    temp = trianglexy[skipStart:-skipEnd,:]
    x, y = line.get_data()
    # print(temp[:10,:] )
    # print(line._xy )
    test2 = np.all(y - temp[:10,1] == 0)
    
    test =  np.all([test1, test2])
    
    assert test
    






import numpy as np
import hysteresis as hys
import matplotlib.pyplot as plt

from hysteresis.resample import (_getBreakPoints, _getOutsideRegions,
                                 resampleRegion)

t = np.linspace(0,4,1000)*np.pi
x = np.sin(t)
y = np.cos(t)*t
xy = np.column_stack([x,y])
ty = np.column_stack([t ,y])
tyReversed = np.column_stack([t[::-1] ,y[::-1]+y])



t2 = np.linspace(0,4,100)*np.pi
x2 = np.sin(t2)
y2 = np.cos(t2)*t2
xy = np.column_stack([x2,y2])
ty2 = np.column_stack([t2 ,y2])
tyReversed = np.column_stack([t2[::-1] ,y2[::-1]+y2])

tt = np.column_stack([t2 ,t2])
tHys = hys.Hysteresis(np.concatenate([ty2, tyReversed]))
# test.plot()
tHys.plot(marker='.')
resampled = hys.resampleRegion(tHys,20)
resampled.cycles[0].plot(marker='.', linewidth=0)
tHys.cycles[0].plot(marker='.', linewidth=0)
len(resampled)


regions1 = [[0.2,0.3], [0.8, 0.9]]
regions2 = [[0.,0.2], [0.8, 1]]

def test_getBreakpoints():
    breakpoints1 = np.array(_getBreakPoints(regions1))
    breakpoints2 = np.array(_getBreakPoints(regions2))
    check1 = np.sum(np.array([0.2, 0.3, 0.8, 0.9]) - breakpoints1) == 0
    check2 = np.sum(np.array([0., 0.2, 0.8, 1]) - breakpoints2) == 0
    assert check1 and check2

def test_getOutsideRegions():
    breakpoints1,_ = _getOutsideRegions(regions1)
    breakpoints2,_ = _getOutsideRegions(regions2)
    breakpoints1 = np.array(breakpoints1)
    breakpoints2 = np.array(breakpoints2)
    check1 = np.sum(np.array([[0, 0.2], [0.2, 0.3], [0.3, 0.8], [0.8, 0.9], [0.9, 1]]) - breakpoints1) == 0
    check2 = np.sum(np.array([[0, 0.2], [0.2, 0.8], [0.8, 1]]) - breakpoints2) == 0
    assert check1 and check2

"""
I'm having trouble thinking of a way to test these properly..
For now, I'm just checking that the number of indexes changed
correctly as a proxy. Not great.
"""

def test_resampleRegion_length():
    xy = hys.resampleRegion(tt,20)

    myHys = hys.Hysteresis(xy)
    assert len(myHys) == 120

def test_resampleRegion_length2():

    xy = hys.resampleRegion(ty, 5, regions1)
    myHys = hys.Hysteresis(xy)
    assert len(myHys) == 808
     
def test_resampleRegion_length3():
    xy = hys.resampleRegion(ty, 5, regions2)
    myHys = hys.Hysteresis(xy)
    # myHys.plot(marker='.')
    assert len(myHys) == 610
    


def test_resample_Monotonic():
    
    curve = hys.MonotonicCurve(tt)
    resampled = hys.resampleRegion(curve,20)

    assert len(resampled) == 120    
    
def test_resample_SimpleCurve_NoSubCycles():
    
    curve = hys.SimpleCurve(tt)
    resampled = hys.resampleRegion(curve,20)

    assert len(resampled) == 120    

def test_resample_SimpleCurve_SubCycles():
    """
    Confirms that the subcycle is resampled properly!
    """
    
    myHys = hys.SimpleCurve(ty2,True)
    # curve = hys.SimpleCurve(tt)
    resampled = hys.resampleRegion(myHys,20)
       
    assert len(resampled.subCycles[0]) == 46    


def test_resample_Hysteresis():
    """
    This is a test by proxy - I've manually looked a the curve to see
    I'm happy with it, and now we just confirm it's the same.
    Again, not great.

    """
    
    
    resampled = hys.resampleRegion(tHys,20)
       
    assert len(resampled) == 239   



# -*- coding: utf-8 -*-
"""
Created on Sat May  1 23:49:09 2021

@author: Christian
"""

import numpy as np
import hysteresis as hys
from  scipy.signal import sawtooth

t = np.linspace(0,5*np.pi,1001)
y = np.sin(t)
xy = np.column_stack([t,y])

# triangleSmall = sawtooth(t *20,0.5)/7
# trianglexy = np.column_stack((triangleSmall, t))

# myHys = hys.Hysteresis(trianglexy)
# myHys.setLength()
# myHys.recalculateCycles_dist()
# myHys.plot(True)
# myHys.re

def test_recalc():
    newHys = hys.SimpleCurve(xy)
    newHys.setPeaks()
    newHys.recalculatePeaks(peakProminence=3)
    assert len(newHys.peakIndexes) == 2


def test_recalc_dist():
    
    newHys = hys.SimpleCurve(xy)
    newHys.setPeaks()
    newHys.recalculatePeaks(peakProminence=3)
    assert len(newHys.peakIndexes) == 2

def test_recalc_like():
    revDist =2
    revWidth = 10
    revProminence = 50
        
    myHys = hys.Hysteresis(xy,revDist,revWidth,revProminence)
    
    newHys = hys.Hysteresis(xy)
    
    newHys.recalculateCycles_like(myHys)
    
    
    check1 = newHys.revDist == revDist
    check2 = newHys.revWidth == revWidth
    check3 = newHys.revProminence == revProminence
    
    assert np.all([check1, check2, check3])


test_recalc_like()
# print(myHys.peakProminence)

# out = myHys + t